# -*- coding: utf-8 -*-
"""
DisplayCAL setup.py script

Can be used with setuptools or pure distutils (the latter can be forced
with the --use-distutils option, otherwise it will try to use setuptools
by default).

Also supported in addition to standard distutils/setuptools commands,
are the bdist_bbfreeze, py2app and py2exe commands (if the appropriate
packages are installed), which makes this file your all-around building/
bundling powerhouse for DisplayCAL. In the case of py2exe, special care
is taken of Python 2.6+ and the Microsoft.VC90.CRT assembly dependency,
so if building an executable on Windows with Python 2.6+ you should
preferably use py2exe. Please note that bdist_bbfreeze and py2app
*require* setuptools.

IMPORTANT NOTE:
If called from within the installed package, should only be used to
uninstall (setup.py uninstall --record=INSTALLED_FILES), otherwise use
the wrapper script in the root directory of the source tar.gz/zip

"""
import codecs
import ctypes.util
from configparser import ConfigParser
from distutils.command.install import install
import distutils.core
from distutils.util import change_root, get_platform
from fnmatch import fnmatch
import functools
import os
import platform
import re
import shutil
import subprocess as sp
import sys
from time import strftime


# Borrowed from setuptools
def _find_all_simple(path):
    """
    Find all files under 'path'
    """
    results = (
        os.path.join(base, file)
        for base, dirs, files in os.walk(path, followlinks=True)
        for file in files
    )
    return filter(os.path.isfile, results)


def findall(dir=os.curdir):
    """
    Find all files under 'dir' and return the list of full filenames.
    Unless dir is '.', return full filenames with dir prepended.
    """
    files = _find_all_simple(dir)
    if dir == os.curdir:
        make_rel = functools.partial(os.path.relpath, start=dir)
        files = map(make_rel, files)
    return list(files)


import distutils.filelist

distutils.filelist.findall = findall  # Fix findall bug in distutils


from DisplayCAL.defaultpaths import autostart, autostart_home
from DisplayCAL.meta import (
    appstream_id,
    author,
    author_ascii,
    author_email,
    description,
    development_home_page,
    DOMAIN,
    longdesc,
    name,
    py_maxversion,
    py_minversion,
    script2pywname,
    version,
    version_tuple,
    wx_minversion,
)
from DisplayCAL.util_list import intlist
from DisplayCAL.util_os import getenvu, relpath, safe_glob
from DisplayCAL.util_str import safe_str

appname = name

bits = platform.architecture()[0][:2]
pypath = os.path.abspath(__file__)
pydir = os.path.dirname(pypath)
source_dir = os.path.dirname(pydir)


if sys.platform in ("darwin", "win32"):
    # Adjust PATH so ctypes.util.find_library can find SDL2 DLLs (if present)
    pth = getenvu("PATH")
    libpth = os.path.join(pydir, "lib")
    if not pth.startswith(libpth + os.pathsep):
        pth = libpth + os.pathsep + pth
        os.environ["PATH"] = safe_str(pth)

config = {
    "data": ["tests/*.icc"],
    "doc": [
        "CHANGES.html",
        "LICENSE.txt",
        "README.html",
        "README-fr.html",
        "screenshots/*.png",
        "theme/*.png",
        "theme/*.css",
        "theme/*.js",
        "theme/*.svg",
        "theme/icons/favicon.ico",
        "theme/slimbox2/*.css",
        "theme/slimbox2/*.js",
    ],
    # Excludes for .app/.exe builds
    # numpy.lib.utils imports pydoc, which imports Tkinter, but
    # numpy.lib.utils is not even used by DisplayCAL, so omit all
    # Tk stuff
    # Use pyglet with OpenAL as audio backend. We only need
    # pyglet, pyglet.app and pyglet.media
    "excludes": {
        "all": [
            "Tkconstants",
            "Tkinter",
            "pygame",
            "pyglet.canvas",
            "pyglet.extlibs",
            "pyglet.font",
            "pyglet.gl",
            "pyglet.graphics",
            "pyglet.image",
            "pyglet.input",
            "pyglet.text",
            "pyglet.window",
            "pyo",
            "setuptools",
            "tcl",
            "test",
            "yaml",
        ],
        "darwin": ["gdbm"],
        "win32": ["gi", "win32com.client.genpy"],
    },
    "package_data": {
        name: [
            "beep.wav",
            "camera_shutter.wav",
            "ColorLookupTable.fx",
            "lang/*.yaml",
            "linear.cal",
            "pnp.ids",
            "presets/*.icc",
            "quirk.json",
            "ref/*.cie",
            "ref/*.gam",
            "ref/*.icm",
            "ref/*.ti1",
            "report/*.css",
            "report/*.html",
            "report/*.js",
            "test.cal",
            "theme/*.png",
            "theme/*.wav",
            "theme/icons/10x10/*.png",
            "theme/icons/16x16/*.png",
            "theme/icons/32x32/*.png",
            "theme/icons/48x48/*.png",
            "theme/icons/72x72/*.png",
            "theme/icons/128x128/*.png",
            "theme/icons/256x256/*.png",
            "theme/icons/512x512/*.png",
            "theme/jet_anim/*.png",
            "theme/patch_anim/*.png",
            "theme/splash_anim/*.png",
            "theme/shutter_anim/*.png",
            "ti1/*.ti1",
            "x3d-viewer/*.css",
            "x3d-viewer/*.html",
            "x3d-viewer/*.js",
            "xrc/*.xrc",
        ]
    },
    "xtra_package_data": {name: {"win32": [f"theme/icons/{name}-uninstall.ico"]}},
}


def add_lib_excludes(key, excludebits):
    for exclude in excludebits:
        config["excludes"][key].extend([f"{name}.lib{exclude}", f"lib{exclude}"])

    for exclude in ("32", "64"):
        for pycompat in ("38", "39", "310", "311", "312", "313"):
            if key == "win32" and (
                pycompat == str(sys.version_info[0]) + str(sys.version_info[1])
                or exclude == excludebits[0]
            ):
                continue
            config["excludes"][key].extend(
                [
                    f"{name}.lib{exclude}.python{pycompat}",
                    f"{name}.lib{exclude}.python{pycompat}.RealDisplaySizeMM",
                ]
            )


add_lib_excludes("darwin", ["64" if bits == "32" else "32"])
add_lib_excludes("win32", ["64" if bits == "32" else "32"])

msiversion = ".".join(
    (
        str(version_tuple[0]),
        str(version_tuple[1]),
        str(version_tuple[2]),
    )
)

plist_dict = {
    "CFBundleDevelopmentRegion": "English",
    "CFBundleExecutable": name,
    "CFBundleGetInfoString": version,
    "CFBundleIdentifier": ".".join(reversed(DOMAIN.split("."))) + "." + name,
    "CFBundleInfoDictionaryVersion": "6.0",
    "CFBundleLongVersionString": version,
    "CFBundleName": name,
    "CFBundlePackageType": "APPL",
    "CFBundleShortVersionString": version,
    "CFBundleSignature": "????",
    "CFBundleVersion": ".".join(map(str, version_tuple)),
    "NSHumanReadableCopyright": f"© {strftime('%Y')} {author}",
    "LSMinimumSystemVersion": "10.6.0",
}


class Target:
    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)


def create_app_symlinks(dist_dir, scripts):
    maincontents_rel = os.path.join(f"{name}.app", "Contents")
    # Create ref, tests, ReadMe and license symlinks in directory
    # containing the app bundle
    for src, tgt in [
        ("ref", "Reference"),
        ("tests", "Tests"),
        ("CHANGES.html", "CHANGES.html"),
        ("README.html", "README.html"),
        ("README-fr.html", "README-fr.html"),
        ("LICENSE.txt", "LICENSE.txt"),
    ]:
        tgt = os.path.join(dist_dir, tgt)
        if os.path.islink(tgt):
            os.unlink(tgt)
        os.symlink(os.path.join(maincontents_rel, "Resources", src), tgt)
    # Create standalone tools app bundles by symlinking to the main bundle
    scripts = [(script2pywname(script), desc) for script, desc in scripts]
    toolscripts = [
        script for script in [script for script, desc in scripts] if script != name
    ]
    for script, desc in scripts:
        if script in (
            name,
            f"{name}-apply-profiles",
            f"{name}-eeColor-to-madVR-converter",
        ) or script.endswith("-console"):
            continue
        toolname = desc.replace(name, "").strip()
        toolapp = os.path.join(dist_dir, f"{toolname}.app")
        if os.path.isdir(toolapp):
            if (
                input(
                    f'WARNING: The output directory "{toolapp}" and ALL ITS '
                    "CONTENTS will be REMOVED! Continue? (y/n)"
                ).lower()
                == "y"
            ):
                print("Removing dir", toolapp)
                shutil.rmtree(toolapp)
            else:
                raise SystemExit("User aborted")
        toolscript = os.path.join(dist_dir, maincontents_rel, "MacOS", script)
        has_tool_script = os.path.exists(toolscript)
        if not has_tool_script:
            # Don't symlink, apps won't be able to run in parallel!
            shutil.copy(
                os.path.join(dist_dir, maincontents_rel, "MacOS", appname), toolscript
            )
        toolcontents = os.path.join(toolapp, "Contents")
        os.makedirs(toolcontents)
        subdirs = ["Frameworks", "Resources"]
        if has_tool_script:
            # PyInstaller
            subdirs.append("MacOS")
        for entry in os.listdir(os.path.join(dist_dir, maincontents_rel)):
            if entry in subdirs:
                os.makedirs(os.path.join(toolcontents, entry))
                for subentry in os.listdir(
                    os.path.join(dist_dir, maincontents_rel, entry)
                ):
                    src = os.path.join(dist_dir, maincontents_rel, entry, subentry)
                    tgt = os.path.join(toolcontents, entry, subentry)
                    if subentry == "main.py":
                        # py2app
                        with open(src, "r") as main_in:
                            py = main_in.read()
                        py = py.replace("main()", "main(%r)" % script[len(name) + 1 :])
                        with open(tgt, "wb") as main_out:
                            main_out.write(py.encode())
                        continue
                    if subentry == name + ".icns":
                        shutil.copy(
                            os.path.join(pydir, "theme", "icons", f"{script}.icns"),
                            os.path.join(toolcontents, entry, f"{script}.icns"),
                        )
                        continue
                    if subentry == script:
                        # PyInstaller
                        os.rename(src, tgt)
                    elif subentry not in toolscripts:
                        os.symlink(
                            os.path.join(
                                "..", "..", "..", maincontents_rel, entry, subentry
                            ),
                            tgt,
                        )
            elif entry == "Info.plist":
                with codecs.open(
                    os.path.join(dist_dir, maincontents_rel, entry), "r", "UTF-8"
                ) as info_in:
                    infoxml = info_in.read()
                # CFBundleName / CFBundleDisplayName
                infoxml = re.sub(
                    r"(Name</key>\s*<string>)%s" % name,
                    lambda match: match.group(1) + toolname,
                    infoxml,
                )
                # CFBundleIdentifier
                infoxml = infoxml.replace(f".{name}</string>", f".{script}</string>")
                # CFBundleIconFile
                infoxml = infoxml.replace(
                    f"{name}.icns</string>", f"{script}.icns</string>"
                )
                # CFBundleExecutable
                infoxml = re.sub(
                    r"(Executable</key>\s*<string>)%s" % name,
                    lambda match: match.group(1) + script,
                    infoxml,
                )
                with codecs.open(
                    os.path.join(toolcontents, entry), "w", "UTF-8"
                ) as info_out:
                    info_out.write(infoxml)
            else:
                os.symlink(
                    os.path.join("..", "..", maincontents_rel, entry),
                    os.path.join(toolcontents, entry),
                )


def get_data(tgt_dir, key, pkgname=None, subkey=None, excludes=None):
    """Return configured data files.

    Args:
        tgt_dir (str): The target directory.
        key (str): The config key.
        pkgname (Union[None, str]): Name of the package. Default is None.
        subkey (Union[None, str]): Name of the subkey. Default is None.
        excludes (Union[None, List[str]]): List of files to exclude. Default is None.

    Returns:
        List[str]: List of strings showing the paths of the data files.
    """
    files = config[key]
    src_dir = source_dir
    if pkgname:
        files = files[pkgname]
        # modifying the src_dir is not working with py2app, so disabling it.
        # src_dir = os.path.join(src_dir, pkgname)
        if subkey:
            if subkey in files:
                files = files[subkey]
            else:
                files = []
    data = []
    for pth in files:
        if not [exclude for exclude in excludes or [] if fnmatch(pth, exclude)]:
            normalized_path = os.path.normpath(
                os.path.join(tgt_dir, os.path.dirname(pth))
            )
            safe_path = [
                relpath(p, src_dir) for p in safe_glob(os.path.join(src_dir, pth))
            ]
            if pkgname:
                # try looking for the "{src_dir}/{pkgname}/{pth}" too
                safe_path += [
                    relpath(p, src_dir)
                    for p in safe_glob(os.path.join(src_dir, pkgname, pth))
                ]
            data.append((normalized_path, safe_path))
    return data


def get_scripts(excludes=None):
    # It is required that each script has an accompanying .desktop file
    scripts_with_desc = []
    scripts = safe_glob(os.path.join(pydir, "..", "scripts", appname.lower() + "*"))

    def sortbyname(a, b):
        a, b = [os.path.splitext(v)[0] for v in (a, b)]
        if a > b:
            return 1
        elif a < b:
            return -1
        else:
            return 0

    import functools

    scripts = sorted(scripts, key=functools.cmp_to_key(sortbyname))
    for script in scripts:
        script = os.path.basename(script)
        if script == appname.lower() + "-apply-profiles-launcher":
            continue
        desktopfile = os.path.join(pydir, "..", "misc", f"{script}.desktop")
        if os.path.isfile(desktopfile):
            cfg = ConfigParser()
            cfg.read(desktopfile)
            script = cfg.get("Desktop Entry", "Exec").split()[0]
            desc = cfg.get("Desktop Entry", "Name")
        else:
            desc = ""
        if not [exclude for exclude in excludes or [] if fnmatch(script, exclude)]:
            scripts_with_desc.append((script, desc))
    return scripts_with_desc


def setup():
    print("***", os.path.abspath(sys.argv[0]), " ".join(sys.argv[1:]))

    bdist_bbfreeze = "bdist_bbfreeze" in sys.argv[1:]
    bdist_dumb = "bdist_dumb" in sys.argv[1:]
    bdist_win = "bdist_msi" in sys.argv[1:] or "bdist_wininst" in sys.argv[1:]
    debug = 0
    do_full_install = False
    do_install = False
    do_py2app = "py2app" in sys.argv[1:]
    do_py2exe = "py2exe" in sys.argv[1:]
    do_uninstall = "uninstall" in sys.argv[1:]
    doc_layout = "deb" if os.path.exists("/etc/debian_version") else ""
    dry_run = "-n" in sys.argv[1:] or "--dry-run" in sys.argv[1:]
    help = False
    install_data = None  # data files install path (only if given)
    is_rpm_build = "bdist_rpm" in sys.argv[1:] or os.path.abspath(sys.argv[0]).endswith(
        os.path.join(
            os.path.sep,
            "rpm",
            "BUILD",
            f"{name}-{version}",
            os.path.basename(os.path.abspath(sys.argv[0])),
        )
    )
    prefix = ""
    recordfile_name = None  # record installed files to this file
    sdist = "sdist" in sys.argv[1:]
    setuptools = None
    skip_postinstall = "--skip-postinstall" in sys.argv[1:]
    use_distutils = not bdist_bbfreeze and not do_py2app
    use_setuptools = (
        not use_distutils
        or "--use-setuptools" in sys.argv[1:]
        or (os.path.exists("use-setuptools") and "--use-distutils" not in sys.argv[1:])
    )
    use_sdl = "--use-sdl" in sys.argv[1:]

    sys.path.insert(1, os.path.join(os.path.dirname(pydir), "util"))

    current_findall = distutils.filelist.findall

    if use_setuptools:
        if "--use-setuptools" in sys.argv[1:] and not os.path.exists("use-setuptools"):
            open("use-setuptools", "w").close()
        try:
            import setuptools
            from setuptools import setup, Extension, find_packages

            setuptools = True
            print("using setuptools")
            current_findall = find_packages
        except ImportError:
            pass
    else:
        if os.path.exists("use-setuptools"):
            os.remove("use-setuptools")

    if distutils.filelist.findall is current_findall:
        # Fix traversing unneeded dirs which can take a long time (minutes)
        def findall(
            dir=os.curdir,
            original=distutils.filelist.findall,
            listdir=os.listdir,
            basename=os.path.basename,
        ):
            os.listdir = lambda path: [
                entry
                for entry in listdir(path)
                if entry not in ("build", "dist") and not entry.startswith(".")
            ]
            try:
                return original(dir)
            finally:
                os.listdir = listdir

        distutils.filelist.findall = findall

    if not setuptools:
        from distutils.core import setup, Extension

        print("using distutils")

    if do_py2exe:
        setup_do_py2exe()

    if do_uninstall:
        i = sys.argv.index("uninstall")
        sys.argv = sys.argv[:i] + ["install"] + sys.argv[i + 1 :]
        install.create_home_path = lambda self: None

    if (
        skip_instrument_conf_files := "--skip-instrument-configuration-files"
        in sys.argv[1:]
    ):
        i = sys.argv.index("--skip-instrument-configuration-files")
        sys.argv = sys.argv[:i] + sys.argv[i + 1 :]

    if not is_rpm_build:
        skip_instrument_conf_files = True

    if skip_postinstall:
        i = sys.argv.index("--skip-postinstall")
        sys.argv = sys.argv[:i] + sys.argv[i + 1 :]

    if "--use-distutils" in sys.argv[1:]:
        i = sys.argv.index("--use-distutils")
        sys.argv = sys.argv[:i] + sys.argv[i + 1 :]

    if "--use-setuptools" in sys.argv[1:]:
        i = sys.argv.index("--use-setuptools")
        sys.argv = sys.argv[:i] + sys.argv[i + 1 :]

    argv = list(sys.argv[1:])
    for i, arg in enumerate(reversed(argv)):
        n = len(sys.argv) - i - 1
        if arg in (
            "install",
            "install_lib",
            "install_headers",
            "install_scripts",
            "install_data",
        ):
            if arg == "install":
                do_full_install = True
            do_install = True
        elif arg == "-d" and len(sys.argv[1:]) > i:
            dist_dir = sys.argv[i + 2]
        else:
            arg = arg.split("=")
            if arg[0] == "--debug":
                debug = 1 if len(arg) == 1 else int(arg[1])
                sys.argv = sys.argv[:n] + sys.argv[n + 1 :]
            elif len(arg) == 2:
                if arg[0] == "--dist-dir":
                    dist_dir = arg[1]
                elif arg[0] == "--doc-layout":
                    doc_layout = arg[1]
                    sys.argv = sys.argv[:n] + sys.argv[n + 1 :]
                elif arg[0] == "--install-data":
                    install_data = arg[1]
                elif arg[0] == "--prefix":
                    prefix = arg[1]
                elif arg[0] == "--record":
                    recordfile_name = arg[1]
            elif arg[0] == "-h" or arg[0].startswith("--help"):
                help = True

    if not recordfile_name and (do_full_install or do_uninstall):
        recordfile_name = "INSTALLED_FILES"
    # if not do_uninstall:
    # sys.argv.append("--record=" + "INSTALLED_FILES")

    if sys.platform in ("darwin", "win32") or "bdist_egg" in sys.argv[1:]:
        doc = data = "." if do_py2app or do_py2exe or bdist_bbfreeze else name
    else:
        # Linux/Unix
        data = name
        if doc_layout.startswith("deb"):
            doc = os.path.join("doc", name.lower())
        elif "suse" in doc_layout:
            doc = os.path.join("doc", "packages", name)
        else:
            doc = os.path.join("doc", f"{name}-{version}")

        if not install_data:
            data = os.path.join("share", data)
            doc = os.path.join("share", doc)
            if is_rpm_build:
                doc = os.path.join(os.path.sep, "usr", doc)

    # Use CA file from certifi project
    if do_py2app or do_py2exe:
        import certifi

        cacert = certifi.where()
        if cacert:
            shutil.copyfile(cacert, os.path.join(pydir, "cacert.pem"))
            config["package_data"][name].append("cacert.pem")
        else:
            print("WARNING: cacert.pem from certifi project not found!")

    # on Mac OS X and Windows, we want data files in the package dir
    # (package_data will be ignored when using py2exe)
    package_data = {
        name: (
            config["package_data"][name]
            if sys.platform in ("darwin", "win32") and not do_py2app and not do_py2exe
            else []
        )
    }
    if sdist and sys.platform in ("darwin", "win32"):
        package_data[name].extend(
            ["theme/icons/22x22/*.png", "theme/icons/24x24/*.png"]
        )
    if sys.platform == "win32" and not do_py2exe:
        package_data[name].append("theme/icons/*.ico")
    # Scripts
    if sys.platform == "darwin":
        scripts = get_scripts(excludes=[appname.lower() + "-apply-profiles"])
    else:
        scripts = get_scripts()
    # Doc files
    data_files = []
    if not is_rpm_build or doc_layout.startswith("deb"):
        data_files += get_data(doc, "doc", excludes=["LICENSE.txt"])
    if data_files:
        if doc_layout.startswith("deb"):
            data_files.append((doc, [os.path.join(pydir, "..", "dist", "copyright")]))
            data_files.append(
                (
                    os.path.join(os.path.dirname(data), "doc-base"),
                    [os.path.join(pydir, "..", "misc", appname.lower() + "-readme")],
                )
            )
        else:
            data_files.append(
                (doc, [relpath(os.path.join(pydir, "..", "LICENSE.txt"), source_dir)])
            )

    # metainfo / appdata.xml
    data_files.append(
        (
            os.path.join(os.path.dirname(data), "metainfo"),
            [
                relpath(
                    os.path.normpath(
                        os.path.join(pydir, "..", "dist", f"{appstream_id}.appdata.xml")
                    ),
                    source_dir,
                )
            ],
        )
    )

    if sys.platform not in ("darwin", "win32") or do_py2app or do_py2exe:
        # Linux/Unix or py2app/py2exe
        data_files += get_data(data, "package_data", name, excludes=["theme/icons/*"])
        data_files += get_data(data, "data")
        data_files += get_data(data, "xtra_package_data", name, sys.platform)
        if sys.platform == "win32":
            # Add python and pythonw
            data_files.extend(
                [
                    (
                        os.path.join(data, "lib"),
                        [
                            sys.executable,
                            os.path.join(
                                os.path.dirname(sys.executable), "pythonw.exe"
                            ),
                        ],
                    )
                ]
            )
            if use_sdl:
                # SDL DLLs for audio module
                sdl2 = ctypes.util.find_library("SDL2")
                sdl2_mixer = ctypes.util.find_library("SDL2_mixer")
                if sdl2:
                    sdl2_libs = [sdl2]
                    if sdl2_mixer:
                        sdl2_libs.append(sdl2_mixer)
                        data_files.append((os.path.join(data, "lib"), sdl2_libs))
                        config["excludes"]["all"].append("pyglet")
                    else:
                        print("WARNING: SDL2_mixer not found!")
                else:
                    print("WARNING: SDL2 not found!")
            if "pyglet" not in config["excludes"]["all"]:
                # OpenAL DLLs for pyglet
                openal32 = ctypes.util.find_library("OpenAL32.dll")
                wrap_oal = ctypes.util.find_library("wrap_oal.dll")
                if openal32:
                    oal = [openal32]
                    if wrap_oal:
                        oal.append(wrap_oal)
                    else:
                        print("WARNING: wrap_oal.dll not found!")
                    data_files.append((data, oal))
                else:
                    print("WARNING: OpenAL32.dll not found!")
        elif sys.platform != "darwin":
            # Linux
            data_files.append(
                (
                    os.path.join(os.path.dirname(data), "applications"),
                    [os.path.join(pydir, "..", "misc", f"{name.lower()}.desktop")]
                    + safe_glob(
                        os.path.join(pydir, "..", "misc", f"{name.lower()}-*.desktop")
                    ),
                )
            )
            data_files.append(
                (
                    (
                        autostart
                        if os.geteuid() == 0 or prefix.startswith("/")
                        else autostart_home
                    ),
                    [
                        os.path.join(
                            pydir,
                            "..",
                            "misc",
                            f"z-{name.lower()}-apply-profiles.desktop",
                        )
                    ],
                )
            )
            data_files.append(
                (
                    os.path.join(os.path.dirname(data), "man", "man1"),
                    safe_glob(os.path.join(pydir, "..", "man", "*.1")),
                )
            )
            if not skip_instrument_conf_files:
                # device configuration / permission stuff
                if is_rpm_build:
                    # RPM postinstall script will install these to the correct
                    # locations. This allows us compatibility with Argyll
                    # packages which may also contain same udev rules / hotplug
                    # scripts, thus avoiding file conflicts
                    data_files.append(
                        (
                            os.path.join(data, "usb"),
                            [os.path.join(pydir, "..", "misc", "45-Argyll.rules")],
                        )
                    )
                    data_files.append(
                        (
                            os.path.join(data, "usb"),
                            [os.path.join(pydir, "..", "misc", "55-Argyll.rules")],
                        )
                    )
                    data_files.append(
                        (
                            os.path.join(data, "usb"),
                            [os.path.join(pydir, "..", "misc", "Argyll")],
                        )
                    )
                    data_files.append(
                        (
                            os.path.join(data, "usb"),
                            [os.path.join(pydir, "..", "misc", "Argyll.usermap")],
                        )
                    )
                else:
                    devconf_files = []
                    if os.path.isdir("/etc/udev/rules.d"):
                        if safe_glob("/dev/bus/usb/*/*"):
                            # USB and serial instruments using udev, where udev
                            # already creates /dev/bus/usb/00X/00X devices
                            devconf_files.append(
                                (
                                    "/etc/udev/rules.d",
                                    [
                                        os.path.join(
                                            pydir, "..", "misc", "55-Argyll.rules"
                                        )
                                    ],
                                )
                            )
                        else:
                            # USB using udev, where there are NOT /dev/bus/usb/00X/00X
                            # devices
                            devconf_files.append(
                                (
                                    "/etc/udev/rules.d",
                                    [
                                        os.path.join(
                                            pydir, "..", "misc", "45-Argyll.rules"
                                        )
                                    ],
                                )
                            )
                    else:
                        if os.path.isdir("/etc/hotplug"):
                            # USB using hotplug and Serial using udev
                            # (older versions of Linux)
                            devconf_files.append(
                                (
                                    "/etc/hotplug/usb",
                                    [
                                        os.path.join(pydir, "..", "misc", fname)
                                        for fname in ["Argyll", "Argyll.usermap"]
                                    ],
                                )
                            )
                    for entry in devconf_files:
                        for fname in entry[1]:
                            if os.path.isfile(fname):
                                data_files.extend([(entry[0], [fname])])
        for dname in (
            "10x10",
            "16x16",
            "22x22",
            "24x24",
            "32x32",
            "48x48",
            "72x72",
            "128x128",
            "256x256",
            "512x512",
        ):
            # Get all the icons needed, depending on platform
            # Only the icon sizes 10, 16, 32, 72, 256 and 512 include icons
            # that are used exclusively for UI elements.
            # These should be installed in an app-specific location, e.g.
            # under Linux $XDG_DATA_DIRS/DisplayCAL/theme/icons/
            # The app icon sizes 16, 32, 48 and 256 (128 under Mac OS X),
            # which are used for taskbar icons and the like, as well as the
            # other sizes can be installed in a generic location, e.g.
            # under Linux $XDG_DATA_DIRS/icons/hicolor/<size>/apps/
            # Generally, icon filenames starting with the lowercase app name
            # should be installed in the generic location.
            icons = []
            desktopicons = []
            if sys.platform == "darwin":
                largest_iconbundle_icon_size = "128x128"
            else:
                largest_iconbundle_icon_size = "256x256"
            for iconpath in safe_glob(
                os.path.join(pydir, "theme", "icons", dname, "*.png")
            ):
                if not os.path.basename(iconpath).startswith(name.lower()) or (
                    sys.platform in ("darwin", "win32")
                    and dname
                    in ("16x16", "32x32", "48x48", largest_iconbundle_icon_size)
                ):
                    # In addition to UI element icons, we also need all the app
                    # icons we use in get_icon_bundle under macOS/Windows,
                    # otherwise they wouldn't be included (under Linux, these
                    # are included for installation to the system-wide icon
                    # theme location instead)
                    icons.append(iconpath)
                elif sys.platform not in ("darwin", "win32"):
                    desktopicons.append(iconpath)
            if icons:
                data_files.append((os.path.join(data, "theme", "icons", dname), icons))
            if desktopicons:
                data_files.append(
                    (
                        os.path.join(
                            os.path.dirname(data), "icons", "hicolor", dname, "apps"
                        ),
                        desktopicons,
                    )
                )
    if do_py2app:
        data_files.append(
            (
                os.path.join(data, "scripts"),
                [
                    os.path.join(
                        pydir,
                        "..",
                        "scripts",
                        f"{name.lower()}-eecolor-to-madvr-converter",
                    )
                ],
            )
        )

    sources = [os.path.join(name, "RealDisplaySizeMM.c")]
    if sys.platform == "win32":
        macros = [("NT", None)]
        libraries = ["user32", "gdi32"]
        link_args = None
    elif sys.platform == "darwin":
        macros = [("__APPLE__", None), ("UNIX", None)]
        libraries = None
        # XXX: Not sure which macOS version exactly removed the need
        # to specify -framework
        if intlist(platform.mac_ver()[0].split(".")) >= [10, 7]:
            link_args = None
        else:
            link_args = [
                "-framework Carbon",
                "-framework CoreFoundation",
                "-framework Python",
                "-framework IOKit",
            ]
            if not help and (
                "build" in sys.argv[1:]
                or "build_ext" in sys.argv[1:]
                or (
                    ("install" in sys.argv[1:] or "install_lib" in sys.argv[1:])
                    and "--skip-build" not in sys.argv[1:]
                )
            ):
                p = sp.Popen(
                    [
                        sys.executable,
                        "-c",
                        f"""import os
from distutils.core import setup, Extension

setup(ext_modules=[Extension("{name}.lib{bits}.RealDisplaySizeMM", sources={sources}, define_macros={macros}, extra_link_args={link_args})])""",
                    ]
                    + sys.argv[1:],
                    stdout=sp.PIPE,
                    stderr=sp.STDOUT,
                )
                lines = []
                while True:
                    o = p.stdout.readline()
                    if o == "" and p.poll() is not None:
                        break
                    if o[0:4] == "gcc ":
                        lines.append(o)
                    print(o.rstrip())
                if len(lines):
                    os.environ["MACOSX_DEPLOYMENT_TARGET"] = "10.5"
                    sp.call(lines[-1], shell=True)  # fix the library
    else:
        macros = [("UNIX", None)]
        libraries = ["X11", "Xinerama", "Xrandr", "Xxf86vm"]
        link_args = None
    if sys.platform == "darwin":
        extname = f"{name}.lib{bits}.RealDisplaySizeMM"
    else:
        extname = f"{name}.lib{bits}.python{sys.version_info[0]}{sys.version_info[1]}.RealDisplaySizeMM"

    RealDisplaySizeMM = Extension(
        extname,
        sources=sources,
        define_macros=macros,
        libraries=libraries,
        extra_link_args=link_args,
    )
    ext_modules = [RealDisplaySizeMM]

    requires = []
    if not setuptools or sys.platform != "win32":
        # wxPython windows installer doesn't add egg-info entry, so
        # a dependency check from pkg_resources would always fail
        requires.append(
            "wxPython (>= {})".format(".".join(str(n) for n in wx_minversion))
        )
    if sys.platform == "win32":
        requires.append("pywin32 (>= 213.0)")

    packages = [name, f"{name}.lib", f"{name}.lib.agw"]
    if sdist:
        # For source distributions we want all libraries
        for pycompat in ("38", "39", "310", "311", "312", "313"):
            packages.extend([f"{name}.lib{bits}", f"{name}.lib{bits}.python{pycompat}"])
    elif sys.platform == "darwin":
        # On Mac OS X we only want the universal binaries
        packages.append(f"{name}.lib{bits}")
    elif sys.platform == "win32":
        # On Windows we want separate libraries
        packages.extend(
            [
                f"{name}.lib{bits}",
                f"{name}.lib{bits}.python{sys.version_info[0]}{sys.version_info[1]}",
            ]
        )
    else:
        # On Linux the libraries will be built on build step
        pass

    attrs = {
        "author": author_ascii,
        "author_email": author_email,
        "classifiers": [
            "Development Status :: 5 - Production/Stable",
            "Environment :: MacOS X",
            "Environment :: Win32 (MS Windows)",
            "Environment :: X11 Applications",
            "Intended Audience :: End Users/Desktop",
            "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
            "Operating System :: OS Independent",
            "Programming Language :: Python :: 3.8",
            "Programming Language :: Python :: 3.9",
            "Programming Language :: Python :: 3.10",
            "Programming Language :: Python :: 3.11",
            "Programming Language :: Python :: 3.12",
            "Programming Language :: Python :: 3.13",
            "Topic :: Multimedia :: Graphics",
        ],
        "data_files": data_files,
        "description": description,
        "download_url": f"{development_home_page}/releases/download/{version}/{name}-{version}.tar.gz",
        "ext_modules": ext_modules,
        "license": "GPL v3",
        "long_description": longdesc,
        "long_description_content_type": "text/x-rst",
        "name": name,
        "packages": packages,
        "package_data": package_data,
        "package_dir": {name: name},
        "platforms": [
            "Python >= {} <= {}".format(
                ".".join(str(n) for n in py_minversion),
                ".".join(str(n) for n in py_maxversion),
            ),
            "Linux/Unix with X11",
            "Mac OS X >= 10.4",
            "Windows 2000 and newer",
        ],
        "requires": requires,
        "provides": [name],
        "scripts": [],
        "url": f"https://{DOMAIN}/",
        "version": msiversion if "bdist_msi" in sys.argv[1:] else version,
    }

    if setuptools:
        attrs["entry_points"] = {
            "gui_scripts": [
                "{} = {}.main:main{}".format(
                    script,
                    name,
                    (
                        ""
                        if script == name.lower()
                        else script[len(name) :].lower().replace("-", "_")
                    ),
                )
                for script, desc in scripts
            ]
        }
        attrs["exclude_package_data"] = {name: ["RealDisplaySizeMM.c"]}
        attrs["include_package_data"] = (
            sys.platform in ("darwin", "win32") and not do_py2app
        )
        install_requires = [req.replace("(", "").replace(")", "") for req in requires]
        attrs["install_requires"] = install_requires
        attrs["zip_safe"] = False
    else:
        attrs["scripts"].extend(
            os.path.join("scripts", script)
            for script, desc in [
                script_desc
                for script_desc in scripts
                if script_desc[0] != f"{name.lower()}-apply-profiles"
                or sys.platform != "darwin"
            ]
        )

    if bdist_bbfreeze:
        attrs["setup_requires"] = ["bbfreeze"]

    if "bdist_wininst" in sys.argv[1:]:
        attrs["scripts"].append(os.path.join("util", f"{name}_postinstall.py"))

    if do_py2app:
        mainpy = os.path.join(source_dir, "main.py")
        if not os.path.exists(mainpy):
            shutil.copy(os.path.join(source_dir, "scripts", name.lower()), mainpy)
        attrs["app"] = [mainpy]
        dist_dir = os.path.join(
            pydir,
            "..",
            "dist",
            f"py2app.{get_platform()}-py{sys.version_info[0]}.{sys.version_info[1]}",
            f"{name}-{version}",
        )
        from py2app.build_app import py2app as py2app_cls

        py2app_cls._copy_package_data = py2app_cls.copy_package_data

        def copy_package_data(self, package, target_dir):
            # Skip package data which is already included as data files
            if package.identifier.split(".")[0] != name:
                self._copy_package_data(package, target_dir)

        py2app_cls.copy_package_data = copy_package_data
        attrs["options"] = {
            "py2app": {
                "argv_emulation": False,
                "dist_dir": dist_dir,
                "excludes": config["excludes"]["all"] + config["excludes"]["darwin"],
                "iconfile": os.path.join(pydir, "theme", "icons", f"{name}.icns"),
                "optimize": 0,
                "plist": plist_dict,
            }
        }
        if use_sdl:
            attrs["options"]["py2app"]["frameworks"] = ["SDL2", "SDL2_mixer"]
        attrs["setup_requires"] = ["py2app"]

    if do_py2exe:
        import wx
        from winmanifest_util import getmanifestxml

        if platform.architecture()[0] == "64bit":
            arch = "amd64"
        else:
            arch = "x86"
        manifest_xml = getmanifestxml(
            os.path.join(
                pydir,
                "..",
                "misc",
                name
                + (
                    f".exe.{arch}.VC90.manifest"
                    if hasattr(sys, "version_info") and sys.version_info[:2] >= (3, 8)
                    else ".exe.manifest"
                ),
            )
        )
        tmp_scripts_dir = os.path.join(source_dir, "build", "temp.scripts")
        if not os.path.isdir(tmp_scripts_dir):
            os.makedirs(tmp_scripts_dir)
        apply_profiles_launcher = (
            f"{appname.lower()}-apply-profiles-launcher",
            f"{appname} Profile Loader Launcher",
        )
        for script, desc in scripts + [apply_profiles_launcher]:
            shutil.copy(
                os.path.join(source_dir, "scripts", script),
                os.path.join(tmp_scripts_dir, script2pywname(script)),
            )
        attrs["windows"] = [
            Target(
                **{
                    "script": os.path.join(tmp_scripts_dir, script2pywname(script)),
                    "icon_resources": [
                        (
                            1,
                            os.path.join(
                                pydir,
                                "theme",
                                "icons",
                                os.path.splitext(os.path.basename(script))[0] + ".ico",
                            ),
                        )
                    ],
                    "other_resources": [(24, 1, manifest_xml)],
                    "copyright": f"© {strftime('%Y')} {author}",
                    "description": desc,
                }
            )
            for script, desc in [
                script_desc1
                for script_desc1 in scripts
                if script_desc1[0] != appname.lower() + "-eecolor-to-madvr-converter"
                and not script_desc1[0].endswith("-console")
            ]
        ]

        # Add profile loader launcher
        attrs["windows"].append(
            Target(
                **{
                    "script": os.path.join(
                        tmp_scripts_dir, script2pywname(apply_profiles_launcher[0])
                    ),
                    "icon_resources": [
                        (
                            1,
                            os.path.join(
                                pydir,
                                "theme",
                                "icons",
                                appname + "-apply-profiles" + ".ico",
                            ),
                        )
                    ],
                    "other_resources": [(24, 1, manifest_xml)],
                    "copyright": f"© {strftime('%Y')} {author}",
                    "description": apply_profiles_launcher[1],
                }
            )
        )

        # Programs that can run with and without GUI
        console_scripts = [f"{name}-VRML-to-X3D-converter"]  # No "-console" suffix!
        for console_script in console_scripts:
            console_script_path = os.path.join(
                tmp_scripts_dir, console_script + "-console"
            )
            if not os.path.isfile(console_script_path):
                shutil.copy(
                    os.path.join(
                        source_dir, "scripts", console_script.lower() + "-console"
                    ),
                    console_script_path,
                )
        attrs["console"] = [
            Target(
                **{
                    "script": os.path.join(
                        tmp_scripts_dir, script2pywname(script) + "-console"
                    ),
                    "icon_resources": [
                        (
                            1,
                            os.path.join(
                                pydir,
                                "theme",
                                "icons",
                                os.path.splitext(os.path.basename(script))[0] + ".ico",
                            ),
                        )
                    ],
                    "other_resources": [(24, 1, manifest_xml)],
                    "copyright": f"© {strftime('%Y')} {author}",
                    "description": desc,
                }
            )
            for script, desc in [
                script_desc2
                for script_desc2 in scripts
                if script2pywname(script_desc2[0]) in console_scripts
            ]
        ]

        # Programs without GUI
        attrs["console"].append(
            Target(
                **{
                    "script": os.path.join(
                        tmp_scripts_dir, appname + "-eeColor-to-madVR-converter"
                    ),
                    "icon_resources": [
                        (
                            1,
                            os.path.join(
                                pydir, "theme", "icons", appname + "-3DLUT-maker.ico"
                            ),
                        )
                    ],
                    "other_resources": [(24, 1, manifest_xml)],
                    "copyright": f"© {strftime('%Y')} {author}",
                    "description": "Convert eeColor 65^3 to madVR 256^3 3D LUT "
                    "(video levels in, video levels out)",
                }
            )
        )

        dist_dir = os.path.join(
            pydir,
            "..",
            "dist",
            f"py2exe.{get_platform()}-py{sys.version_info[0]}.{sys.version_info[1]}",
            f"{name}-{version}",
        )
        attrs["options"] = {
            "py2exe": {
                "dist_dir": dist_dir,
                "dll_excludes": [
                    "iertutil.dll",
                    "MPR.dll",
                    "msvcm90.dll",
                    "msvcp90.dll",
                    "msvcr90.dll",
                    "mswsock.dll",
                    "urlmon.dll",
                    "w9xpopen.exe",
                    "gdiplus.dll",
                    "mfc90.dll",
                ],
                "excludes": config["excludes"]["all"] + config["excludes"]["win32"],
                "bundle_files": 3 if wx.VERSION >= (2, 8, 10, 1) else 1,
                "compressed": 1,
                "optimize": 0,  # 0 = don’t optimize (generate .pyc)
                # 1 = normal optimization (like python -O)
                # 2 = extra optimization (like python -OO)
            }
        }
        if debug:
            attrs["options"]["py2exe"].update(
                {"bundle_files": 3, "compressed": 0, "optimize": 0, "skip_archive": 1}
            )
        if setuptools:
            attrs["setup_requires"] = ["py2exe"]
        attrs["zipfile"] = os.path.join("lib", "library.zip")

    if (do_uninstall or do_install or bdist_win or bdist_dumb) and not help:
        distutils.core._setup_stop_after = "commandline"
        dist = setup(**attrs)
        distutils.core._setup_stop_after = None
        cmd = install(dist).get_finalized_command("install")
        if debug > 0:
            for attrname in [
                "base",
                "data",
                "headers",
                "lib",
                "libbase",
                "platbase",
                "platlib",
                "prefix",
                "purelib",
                "root",
                "scripts",
                "userbase",
            ]:
                if attrname not in ["prefix", "root"]:
                    attrname = "install_" + attrname
                if hasattr(cmd, attrname):
                    print(attrname, getattr(cmd, attrname))
        if debug > 1:
            try:
                from ppdir import ppdir
            except ImportError:
                pass
            else:
                ppdir(cmd, types=[dict, list, str, tuple, type, str])
        if not install_data:
            if sys.platform in ("darwin", "win32"):
                # on Mac OS X and Windows, we want data files in the package dir
                data_basedir = cmd.install_lib
            else:
                data_basedir = cmd.install_data
        else:
            data_basedir = install_data

        data = change_root(data_basedir, data)
        doc = change_root(data_basedir, doc)
        # determine in which cases we want to make data file paths relative to
        # site-packages (on Mac and Windows) and when we want to make them
        # absolute (Linux)
        linux = sys.platform not in ("darwin", "win32") and (
            not cmd.root and setuptools
        )
        dar_win = (
            sys.platform in ("darwin", "win32") and (cmd.root or not setuptools)
        ) or bdist_win
        if (
            not do_uninstall
            and not install_data
            and (linux or dar_win)
            and attrs["data_files"]
        ):
            if data_basedir.startswith(cmd.install_data + os.path.sep):
                data_basedir = relpath(data_basedir, cmd.install_data)
            for i, f in enumerate(attrs["data_files"]):
                if isinstance(f, str):
                    attrs["data_files"][i] = change_root(data_basedir, f)
                else:
                    attrs["data_files"][i] = (change_root(data_basedir, f[0]), f[1])

    if do_uninstall and not help:
        # Quick and dirty uninstall
        if dry_run:
            print("dry run - nothing will be removed")
        else:
            from DisplayCAL.postinstall import postuninstall

            # Yeah, yeah - its actually pre-uninstall
            if cmd.root:
                postuninstall(prefix=change_root(cmd.root, cmd.prefix))
            else:
                postuninstall(prefix=cmd.prefix)

        removed = []
        visited = []

        if os.path.exists(recordfile_name):
            paths = [
                (
                    change_root(cmd.root, line.rstrip("\n"))
                    if cmd.root
                    else line.rstrip("\n")
                )
                for line in open(recordfile_name, "r")
            ]
        else:
            paths = []

        if not paths:
            # If the installed files have not been recorded, use some fallback
            # logic to find them
            paths = safe_glob(os.path.join(cmd.install_scripts, name))
            if sys.platform == "win32":
                if setuptools:
                    paths += safe_glob(os.path.join(cmd.install_scripts, f"{name}.exe"))
                    paths += safe_glob(
                        os.path.join(cmd.install_scripts, f"{name}-script.py")
                    )
                else:
                    paths += safe_glob(os.path.join(cmd.install_scripts, f"{name}.cmd"))
            paths += safe_glob(
                os.path.join(cmd.install_scripts, f"{name}_postinstall.py")
            )
            for attrname in ["data", "headers", "lib", "libbase", "platlib", "purelib"]:
                path = os.path.join(getattr(cmd, f"install_{attrname}"), name)
                if path not in paths:
                    # Using sys.version in this way is consistent with
                    # setuptools
                    paths += (
                        safe_glob(path)
                        + safe_glob(
                            f"{path}-{version}-py{sys.version_info[0]}.{sys.version_info[1]}*.egg"
                        )
                        + safe_glob(
                            f"{path}-{version}-py{sys.version_info[0]}.{sys.version_info[1]}*.egg-info"
                        )
                    )

            if os.path.isabs(data) and data not in paths:
                for fname in [
                    "lang",
                    "presets",
                    "ref",
                    "report",
                    "screenshots",
                    "tests",
                    "theme",
                    "ti1",
                    "x3d-viewer",
                    "CHANGES.html",
                    "LICENSE.txt",
                    "README.html",
                    "README-fr.html",
                    "beep.wav",
                    "cacert.pem",
                    "camera_shutter.wav",
                    "ColorLookupTable.fx",
                    f"{name.lower()}.desktop",
                    f"{name.lower()}-3dlut-maker.desktop",
                    f"{name.lower()}-curve-viewer.desktop",
                    f"{name.lower()}-profile-info.desktop",
                    f"{name.lower()}-scripting-client.desktop",
                    f"{name.lower()}-synthprofile.desktop",
                    f"{name.lower()}-testchart-editor.desktop",
                    "pnp.ids",
                    "quirk.json",
                    "linear.cal",
                    "test.cal",
                ]:
                    path = os.path.join(data, fname)
                    if path not in paths:
                        paths += safe_glob(path)
            if os.path.isabs(doc) and doc not in paths:
                for fname in [
                    "screenshots",
                    "theme",
                    "CHANGES.html",
                    "LICENSE.txt",
                    "README.html",
                    "README-fr.html",
                ]:
                    path = os.path.join(doc, fname)
                    if path not in paths:
                        paths += safe_glob(path)
            if sys.platform == "win32":
                from DisplayCAL.postinstall import get_special_folder_path

                startmenu_programs_common = get_special_folder_path(
                    "CSIDL_COMMON_PROGRAMS"
                )
                startmenu_programs = get_special_folder_path("CSIDL_PROGRAMS")
                for path in (startmenu_programs_common, startmenu_programs):
                    if path:
                        for filename in (
                            name,
                            "CHANGES",
                            "LICENSE",
                            "README",
                            "Uninstall",
                        ):
                            paths += safe_glob(
                                os.path.join(path, name, f"{filename}.lnk")
                            )

        for path in paths:
            if os.path.exists(path):
                if path in visited:
                    continue
                else:
                    visited.append(path)
                if dry_run:
                    print(path)
                    continue
                try:
                    if os.path.isfile(path):
                        os.remove(path)
                    elif os.path.isdir(path):
                        os.rmdir(path)
                except Exception as exception:
                    print("could'nt remove", path)
                    print("   ", exception)
                else:
                    print("removed", path)
                    removed.append(path)
            while path != os.path.dirname(path):
                # remove parent directories if empty
                # could also use os.removedirs(path) but we want some status
                # info
                path = os.path.dirname(path)
                if os.path.isdir(path):
                    if len(os.listdir(path)) == 0:
                        if path in visited:
                            continue
                        else:
                            visited.append(path)
                        if dry_run:
                            print(path)
                            continue
                        try:
                            os.rmdir(path)
                        except Exception as exception:
                            print("could'nt remove", path)
                            print("   ", exception)
                        else:
                            print("removed", path)
                            removed.append(path)
                    else:
                        break

        if not removed:
            print(len(visited), "entries found")
        else:
            print(len(removed), "entries removed")

    else:
        # To have a working sdist and bdist_rpm when using distutils,
        # we go to the length of generating MANIFEST.in from scratch everytime,
        # using the information available from setup.
        manifest_in = ["# This file will be re-generated by setup.py - do not edit"]
        manifest_in.extend(
            [
                "include LICENSE.txt",
                "include MANIFEST",
                "include MANIFEST.in",
                "include README.html",
                "include README-fr.html",
                "include CHANGES.html",
                f"include {name}*.pyw",
                f"include {name}-*.pyw",
                f"include {name}-*.py",
                "include use-distutils",
            ]
        )
        manifest_in.append("include " + os.path.basename(sys.argv[0]))
        manifest_in.append(
            "include " + os.path.splitext(os.path.basename(sys.argv[0]))[0] + ".cfg"
        )
        for _datadir, datafiles in attrs.get("data_files", []):
            for datafile in datafiles:
                manifest_in.append(
                    "include {}".format(
                        relpath(os.path.sep.join(datafile.split("/")), source_dir)
                        or datafile
                    )
                )
        for extmod in attrs.get("ext_modules", []):
            manifest_in.extend(
                "include " + os.path.sep.join(src.split("/")) for src in extmod.sources
            )
        for pkg in attrs.get("packages", []):
            pkg = os.path.join(*pkg.split("."))
            pkgdir = os.path.sep.join(
                attrs.get("package_dir", {}).get(pkg, pkg).split("/")
            )
            manifest_in.append("include " + os.path.join(pkgdir, "*.py"))
            # manifest_in.append("include " + os.path.join(pkgdir, "*.pyd"))
            # manifest_in.append("include " + os.path.join(pkgdir, "*.so"))
            for obj in attrs.get("package_data", {}).get(pkg, []):
                manifest_in.append(
                    "include " + os.path.sep.join([pkgdir] + obj.split("/"))
                )
        for pymod in attrs.get("py_modules", []):
            manifest_in.append("include {}".format(os.path.join(*pymod.split("."))))
        manifest_in.append(
            "include {}".format(os.path.join(name, "theme", "theme-info.txt"))
        )
        manifest_in.append(
            "recursive-include {} {} {}".format(
                os.path.join(name, "theme", "icons"), "*.icns", "*.ico"
            )
        )
        manifest_in.append("include {}".format(os.path.join("man", "*.1")))
        manifest_in.append("recursive-include misc *")
        if skip_instrument_conf_files:
            manifest_in.extend(
                [
                    "exclude misc/Argyll",
                    "exclude misc/*.rules",
                    "exclude misc/*.usermap",
                ]
            )
        manifest_in.append("include {}".format(os.path.join("screenshots", "*.png")))
        manifest_in.append("include {}".format(os.path.join("scripts", "*")))
        manifest_in.append("include {}".format(os.path.join("tests", "*")))
        manifest_in.append("recursive-include theme *")
        manifest_in.append("recursive-include util *.cmd *.py *.sh")
        if sys.platform == "win32" and not setuptools:
            # Only needed under Windows
            manifest_in.append("global-exclude .svn/*")
        manifest_in.append("global-exclude *~")
        manifest_in.append("global-exclude *.backup")
        manifest_in.append("global-exclude *.bak")
        if not dry_run:
            manifest = open("MANIFEST.in", "w")
            manifest.write("\n".join(manifest_in))
            manifest.close()
            if os.path.exists("MANIFEST"):
                os.remove("MANIFEST")

        if bdist_bbfreeze:
            i = sys.argv.index("bdist_bbfreeze")
            if "-d" not in sys.argv[i + 1 :] and "--dist-dir" not in sys.argv[i + 1 :]:
                dist_dir = os.path.join(
                    pydir,
                    "..",
                    "dist",
                    f"bbfreeze.{get_platform()}-py{sys.version_info[0]}.{sys.version_info[1]}",
                )
                sys.argv.insert(i + 1, f"--dist-dir={dist_dir}")
            if "egg_info" not in sys.argv[1:i]:
                sys.argv.insert(i, "egg_info")

        if do_py2app or do_py2exe:
            sys.path.insert(1, pydir)
            i = sys.argv.index("py2app" if do_py2app else "py2exe")
            if "build_ext" not in sys.argv[1:i]:
                sys.argv.insert(i, "build_ext")

        setup(**attrs)

        if dry_run or help:
            return

            import wx

            if wx.VERSION >= (4,):
                # Fix wxPython 4 dylibs being included in wrong location
                wx_dylibs = os.path.join(lib_dynload_dir, "wx")
                for entry in os.listdir(frameworks_dir):
                    if entry.startswith("libwx"):
                        dylib = os.path.join(frameworks_dir, entry)
                        lib_dylib = os.path.join(wx_dylibs, entry)
                        print("Moving", dylib, "->", lib_dylib)
                        shutil.move(dylib, lib_dylib)

            create_app_symlinks(dist_dir, scripts)

        if do_py2exe:
            shutil.copy(
                os.path.join(
                    dist_dir, f"python{sys.version_info[0]}{sys.version_info[1]}.dll"
                ),
                os.path.join(
                    dist_dir,
                    "lib",
                    f"python{sys.version_info[0]}{sys.version_info[1]}.dll",
                ),
            )

        if (
            (bdist_bbfreeze and sys.platform == "win32") or do_py2exe
        ) and sys.version_info[:2] >= (2, 6):
            from vc90crt import name as vc90crt_name, vc90crt_copy_files

            if do_py2exe:
                vc90crt_copy_files(dist_dir)
                vc90crt_copy_files(os.path.join(dist_dir, "lib"))
            else:
                vc90crt_copy_files(os.path.join(dist_dir, name + "-" + version))

        if do_full_install and not is_rpm_build and not skip_postinstall:
            from DisplayCAL.postinstall import postinstall

            if sys.platform == "win32":
                path = os.path.join(cmd.install_lib, name)
                # Using sys.version in this way is consistent with setuptools
                for path in safe_glob(path) + safe_glob(
                    os.path.join(
                        f"{path}-{version}-py{sys.version[:3]}*.egg",
                        name,
                    )
                ):
                    if cmd.root:
                        postinstall(prefix=change_root(cmd.root, path))
                    else:
                        postinstall(prefix=path)

            elif cmd.root:
                postinstall(prefix=change_root(cmd.root, cmd.prefix))
            else:
                postinstall(prefix=cmd.prefix)


def setup_do_py2exe():
    import py2exe

    # ModuleFinder can't handle runtime changes to __path__, but win32com
    # uses them
    try:
        # if this doesn't work, try import modulefinder
        import py2exe.mf as modulefinder
        import win32com

        for p in win32com.__path__[1:]:
            modulefinder.AddPackagePath("win32com", p)
        for extra in ["win32com.shell"]:
            __import__(extra)
            m = sys.modules[extra]
            for p in m.__path__[1:]:
                modulefinder.AddPackagePath(extra, p)
    except ImportError:
        # no build path setup, no worries.
        pass
    origIsSystemDLL = py2exe.build_exe.isSystemDLL
    systemroot = os.getenv("SystemRoot").lower()

    def isSystemDLL(pathname):
        if (
            os.path.basename(pathname).lower() in ("gdiplus.dll", "mfc90.dll")
            or os.path.basename(pathname).lower().startswith("python")
            or os.path.basename(pathname).lower().startswith("pywintypes")
        ):
            return 0
        return pathname.lower().startswith(systemroot + "\\")

    py2exe.build_exe.isSystemDLL = isSystemDLL


if __name__ == "__main__":
    setup()



# -*- coding: utf-8 -*-

import math
import os
import sys

from DisplayCAL.argyll_instruments import get_canonical_instrument_name, instruments
from DisplayCAL.config import getcfg
from DisplayCAL.debughelpers import UnloggedError
from DisplayCAL.meta import name as appname
from DisplayCAL.util_str import make_filename_safe
from DisplayCAL.worker_base import get_argyll_util
from DisplayCAL.wxaddons import wx
from DisplayCAL.wxLUTViewer import LUTCanvas
from DisplayCAL.wxwindows import FlatShadedButton, show_result_dialog
from DisplayCAL import CGATS
from DisplayCAL import colormath
from DisplayCAL import config
from DisplayCAL import ICCProfile as ICCP
from DisplayCAL import localization as lang
from DisplayCAL import wxenhancedplot as plot


BGCOLOUR = "#101010"
FGCOLOUR = "#999999"
GRIDCOLOUR = "#202020"

if sys.platform == "darwin":
    FONTSIZE_LARGE = 11
    FONTSIZE_SMALL = 10
else:
    FONTSIZE_LARGE = 9
    FONTSIZE_SMALL = 8

NTICK = 10


# Graph labeling functions
# See Argyll plot/plot.c


def expt(a, n):
    return math.pow(a, n)


def nicenum(x, do_round):
    if x < 0.0:
        x = -x
    ex = math.floor(math.log10(x))
    f = x / expt(10.0, ex)
    if do_round:
        if f < 1.5:
            nf = 1.0
        elif f < 3.0:
            nf = 2.0
        elif f < 7.0:
            nf = 5.0
        else:
            nf = 10.0
    else:
        if f < 1.0:
            nf = 1.0
        elif f < 2.0:
            nf = 2.0
        elif f < 5.0:
            nf = 5.0
        else:
            nf = 10.0
    return nf * expt(10.0, ex)


class CCXXPlot(wx.Frame):
    """CCMX/CCSS plot and information"""

    def __init__(self, parent, cgats, worker=None):
        """Init new CCXPlot window.

        parent   Parent window (only used for error dialogs)
        cgats    A CCMX/CCSS CGATS instance
        worker   Worker instance

        """
        self.is_ccss = cgats[0].type == b"CCSS"

        desc = cgats.get_descriptor()

        if cgats.filename:
            fn, ext = os.path.splitext(os.path.basename(cgats.filename))
        else:
            fn = desc
            if self.is_ccss:
                ext = ".ccss"
            else:
                ext = ".ccmx"

        if isinstance(fn, bytes):
            fn = fn.decode("utf-8")

        desc = lang.getstr(f"{ext[1:]}.{fn}", default=desc)

        if self.is_ccss:
            ccxx_type = "spectral"
        else:
            ccxx_type = "matrix"

        title = "%s: %s" % (
            lang.getstr(ccxx_type),
            desc if isinstance(desc, str) else desc.decode("utf-8"),
        )

        if self.is_ccss:
            # Convert to TI3 so we can get XYZ from spectra for coloring

            temp = worker.create_tempdir()
            if isinstance(temp, Exception):
                show_result_dialog(temp, parent)
            else:
                basename = make_filename_safe(desc)
                if isinstance(basename, bytes):
                    basename = basename.decode("utf-8")
                temp_path = os.path.join(temp, basename + ".ti3")

                cgats[0].type = b"CTI3"
                cgats[0].DEVICE_CLASS = b"DISPLAY"
                cgats.write(temp_path)

                temp_out_path = os.path.join(temp, basename + ".CIE.ti3")

                result = worker.exec_cmd(
                    get_argyll_util("spec2cie"),
                    [temp_path, temp_out_path],
                    capture_output=True,
                )
                if isinstance(result, Exception) or not result:
                    show_result_dialog(
                        UnloggedError(result or "".join(worker.errors)), parent
                    )
                    worker.wrapup(False)
                else:
                    try:
                        cgats = CGATS.CGATS(temp_out_path)
                    except Exception as exception:
                        show_result_dialog(exception, parent)
                    finally:
                        worker.wrapup(False)

        data_format = cgats.queryv1("DATA_FORMAT")
        data = cgats.queryv1("DATA")

        XYZ_max = 0
        self.samples = []

        if self.is_ccss:
            x_min = cgats.queryv1("SPECTRAL_START_NM")
            x_max = cgats.queryv1("SPECTRAL_END_NM")
            bands = cgats.queryv1("SPECTRAL_BANDS")
            lores = bands <= 40
            if lores:
                # Interpolate if lores
                # 1nm intervals
                steps = int(x_max - x_min) + 1
                print("Up-interpolating", bands, "spectral bands to", steps)
                step = (x_max - x_min) / (steps - 1.0)
            else:
                step = (x_max - x_min) / (bands - 1.0)
            y_min = 0
            y_max = 1

            Y_max = 0
            for i in data:
                sample = data[i]
                # Get nm and spectral power
                values = []
                x = x_min
                for k in data_format.values():
                    if k.startswith(b"SPEC_"):
                        y = sample[k.decode("utf-8")]
                        y_min = min(y, y_min)
                        y_max = max(y, y_max)
                        if lores:
                            values.append(y)
                        else:
                            values.append((x, y))
                            x += step
                if lores:
                    # Interpolate if lores. Use Catmull-Rom instead of
                    # PolySpline as we want curves to go through points exactly
                    numvalues = len(values)
                    interp = ICCP.CRInterpolation(values)
                    values = []
                    for i in range(steps):
                        values.append(
                            (x, interp(i / (steps - 1.0) * (numvalues - 1.0)))
                        )
                        x += step
                # Get XYZ for colorization
                XYZ = []
                for component in "XYZ":
                    label = "XYZ_" + component
                    if label in sample:
                        v = sample[label]
                        XYZ_max = max(XYZ_max, v)
                        if label == "XYZ_Y":
                            Y_max = max(Y_max, v)
                        XYZ.append(v)
                self.samples.append((XYZ, values, {}))

            Plot = plot.PolyLine
            Plot._attributes["width"] = 1
        else:
            # CCMX
            cube_size = 2
            x_min = 0
            y_min = 0

            mtx = colormath.Matrix3x3(
                [
                    [sample[k.decode("utf-8")] for k in data_format.values()]
                    for sample in data.values()
                ]
            )
            imtx = mtx.inverted()

            # Get XYZ that colorimeter would measure without matrix (sRGB ref,
            # so not accurate, but useful for visual representation which is all
            # we care about here)
            if cube_size == 2:
                scale = 1
                x_max = 100 * scale
                y_max = x_max * (74.6 / 67.4)
                if sys.platform != "win32":
                    x_center = x_max / 2.0
                else:
                    x_center = x_max / 2.0 - 2.5
                y_center = y_max / 2.0
                x_center *= scale
                y_center *= scale
                pos2rgb = [
                    ((x_center - 23.7, y_center - 13.7), (0, 0, 1)),
                    ((x_center, y_center + 27.3), (0, 1, 0)),
                    ((x_center + 23.7, y_center - 13.7), (1, 0, 0)),
                    ((x_center - 23.7, y_center + 13.7), (0, 1, 1)),
                    ((x_center, y_center - 27.3), (1, 0, 1)),
                    ((x_center + 23.7, y_center + 13.7), (1, 1, 0)),
                    ((x_center, y_center), (1, 1, 1)),
                ]
                attrs_c = {"size": 10}
                attrs_r = {"size": 5}
            else:
                x_max = 100
                y_max = 100
                y = -5
                pos2rgb = []
                for R in range(cube_size):
                    for G in range(cube_size):
                        x = -5
                        y += 10
                        for B in range(cube_size):
                            x += 10
                            pos2rgb.append(
                                ((x, y), (v / (cube_size - 1.0) for v in (R, G, B)))
                            )
                attrs_c = {"marker": "square", "size": 10}
                attrs_r = {"marker": "square", "size": 5}
            Y_max = (imtx * colormath.get_whitepoint("D65"))[1]
            for _i, ((x, y), (R, G, B)) in enumerate(pos2rgb):
                XYZ = list(colormath.RGB2XYZ(R, G, B))
                X, Y, Z = imtx * XYZ
                XYZ_max = max(XYZ_max, X, Y, Z)
                self.samples.append(([X, Y, Z], [(x, y)], attrs_c))
                self.samples.append((XYZ, [(x, y)], attrs_r))

            Plot = plot.PolyMarker

        if self.is_ccss:
            # Protect against division by zero when range is zero
            if not x_max - x_min:
                x_min = 350.0
                x_max = 750.0
            if not y_max - y_min:
                y_min = 0.0
                y_max = 10.0

            y_zero = 0

            self.ccxx_axis_x = (
                math.floor(x_min / 50.0) * 50,
                math.ceil(x_max / 50.0) * 50,
            )
            self.spec_x = (self.ccxx_axis_x[1] - self.ccxx_axis_x[0]) / 50.0
            graph_range = nicenum(y_max - y_zero, False)
            d = nicenum(graph_range / (NTICK - 1.0), True)
            self.spec_y = math.ceil(y_max / d)
            self.ccxx_axis_y = (math.floor(y_zero / d) * d, self.spec_y * d)
        else:
            self.ccxx_axis_x = (
                math.floor(x_min / 20.0) * 20,
                math.ceil(x_max / 20.0) * 20,
            )
            self.ccxx_axis_y = (math.floor(y_min), math.ceil(y_max))

        self.gfx = []
        for XYZ, values, attrs in self.samples:
            if len(XYZ) == 3:
                # Got XYZ
                if attrs.get("size", 0) > 11.25:
                    # Colorimeter XYZ
                    if Y_max > 1:
                        # Colorimeter brighter than ref
                        XYZ[:] = [v / Y_max for v in XYZ]
                    else:
                        # Colorimeter dimmer than ref
                        XYZ[:] = [v * Y_max for v in XYZ]
                else:
                    # Ref XYZ
                    if Y_max > 1:
                        # Colorimeter brighter than ref
                        XYZ[:] = [v / Y_max for v in XYZ]
                RGB = tuple(
                    int(v) for v in colormath.XYZ2RGB(*XYZ, scale=255, round_=True)
                )
            else:
                RGB = (153, 153, 153)
            self.gfx.append(Plot(values, colour=wx.Colour(*RGB), **attrs))
        if self.is_ccss:
            # Add a few points at the extremes to define a bounding box
            self.gfx.append(
                plot.PolyLine(
                    [
                        (self.ccxx_axis_x[0], self.ccxx_axis_y[0]),
                        (self.ccxx_axis_x[1], self.ccxx_axis_y[1] - y_min),
                    ],
                    colour=wx.Colour(0, 0, 0, 0),
                )
            )

        ref = cgats.queryv1("REFERENCE")
        if ref:
            ref = get_canonical_instrument_name(ref).decode("utf-8")

        if not self.is_ccss:
            observers_ab = {}
            for observer in config.valid_values["observer"]:
                observers_ab[observer] = lang.getstr("observer." + observer)
            x_label = [lang.getstr("matrix")]
            x_label.extend(["%9.6f %9.6f %9.6f" % tuple(row) for row in mtx])
            if ref:
                ref_observer = cgats.queryv1("REFERENCE_OBSERVER")
                if ref_observer:
                    # If the .ccmx does not contain this value,
                    # the graph will still draw
                    ref_observer = ref_observer.decode("utf-8")
                    ref += ", " + observers_ab.get(ref_observer, ref_observer)
                x_label.append("")
                x_label.append(ref)
            fit_method = cgats.queryv1("FIT_METHOD")
            if fit_method == b"xy":
                # If the .ccmx does not contain this value,
                # the graph will still draw
                fit_method = lang.getstr("ccmx.use_four_color_matrix_method")
            elif fit_method:
                fit_method = lang.getstr("perceptual")
            fit_de00_avg = cgats.queryv1("FIT_AVG_DE00")
            if not isinstance(fit_de00_avg, float):
                fit_de00_avg = None
            fit_de00_max = cgats.queryv1("FIT_MAX_DE00")
            if not isinstance(fit_de00_max, float):
                fit_de00_max = None
            if fit_method:
                x_label.append(fit_method)
            fit_de00 = []
            if fit_de00_avg:
                fit_de00.append(
                    "ΔE*00 %s %.4f"
                    % (lang.getstr("profile.self_check.avg"), fit_de00_avg)
                )
            if fit_de00_max:
                fit_de00.append(
                    "ΔE*00 %s %.4f"
                    % (lang.getstr("profile.self_check.max"), fit_de00_max)
                )
            if fit_de00:
                x_label.append("\n".join(fit_de00))
            x_label = "\n".join(x_label)
        else:
            x_label = ""
            if ref:
                x_label += ref + ", "
            x_label += "%.1fnm, %i-%inm" % (
                (x_max - x_min) / (bands - 1.0),
                x_min,
                x_max,
            )

        scale = max(getcfg("app.dpi") / config.get_default_dpi(), 1)

        style = wx.DEFAULT_FRAME_STYLE

        wx.Frame.__init__(self, None, -1, title, style=style)
        self.SetIcons(config.get_icon_bundle([256, 48, 32, 16], appname))
        self.SetBackgroundColour(BGCOLOUR)
        self.Sizer = wx.GridSizer(1, 1, 0, 0)
        bg = wx.Panel(self)
        bg.SetBackgroundColour(BGCOLOUR)
        bg.Sizer = wx.BoxSizer(wx.VERTICAL)
        self.canvas = canvas = LUTCanvas(bg)
        if self.is_ccss:
            bg.MinSize = (int(513 * scale), int(557 * scale))
            btnsizer = wx.BoxSizer(wx.HORIZONTAL)
            bg.Sizer.Add(
                btnsizer,
                flag=wx.EXPAND | wx.TOP,
                border=16,
            )
            self.toggle_btn = FlatShadedButton(bg, -1, label=lang.getstr("spectral"))
            btnsizer.Add(self.toggle_btn, 1)
            self.Sizer.Add(bg, 1, flag=wx.EXPAND)
            bg.Sizer.Add(canvas, 1, flag=wx.EXPAND)
        else:
            bg.MinSize = (int(400 * scale), int(400 * scale))
            self.Sizer.Add(bg, flag=wx.ALIGN_CENTER)
            canvas_w = int(240 * scale)
            canvas.MinSize = (int(canvas_w), int(canvas_w * (74.6 / 67.4)))
            bg.Sizer.Add(canvas, flag=wx.ALIGN_CENTER)
        label = wx.StaticText(
            bg, -1, x_label.replace("&", "&&"), style=wx.ALIGN_CENTRE_HORIZONTAL
        )
        label.SetForegroundColour(FGCOLOUR)
        label.SetMaxFontSize(11)
        bg.Sizer.Add(label, flag=wx.ALIGN_CENTER | wx.ALL & ~wx.TOP, border=16 * scale)
        canvas.SetBackgroundColour(BGCOLOUR)
        canvas.SetEnableCenterLines(False)
        canvas.SetEnableDiagonals(False)
        canvas.SetEnableGrid(True)
        canvas.enableTicks = (True, True)
        canvas.tickPen = wx.Pen(GRIDCOLOUR, int(canvas._pointSize[0]))
        canvas.SetEnablePointLabel(False)
        canvas.SetEnableTitle(True)
        canvas.SetForegroundColour(FGCOLOUR)
        canvas.SetGridColour(GRIDCOLOUR)
        canvas.canvas.BackgroundColour = BGCOLOUR
        if self.is_ccss:
            canvas.HandCursor = wx.StockCursor(wx.CURSOR_SIZING)
            canvas.SetCursor(canvas.HandCursor)
        else:
            canvas.canvas.Unbind(wx.EVT_LEFT_DCLICK)
            canvas.SetEnableDrag(False)
            canvas.SetCursor(wx.StockCursor(wx.CURSOR_DEFAULT))
            canvas.SetXSpec("none")
            canvas.SetYSpec("none")

        # CallAfter is needed under GTK as usual
        wx.CallAfter(self.draw_ccxx)

        if self.is_ccss:
            self.Bind(wx.EVT_KEY_DOWN, self.key_handler)
            for child in self.GetAllChildren():
                child.Bind(wx.EVT_KEY_DOWN, self.key_handler)
                child.Bind(wx.EVT_MOUSEWHEEL, self.OnWheel)

            self.toggle_btn.Bind(wx.EVT_BUTTON, self.toggle_draw)

            self.Bind(wx.EVT_SIZE, self.OnSize)
        else:
            bg.Sizer.Add((0, 16))
        self.Sizer.SetSizeHints(self)
        self.Sizer.Layout()

    def OnSize(self, event):
        if self.canvas.last_draw:
            wx.CallAfter(self.canvas._DrawCanvas, self.canvas.last_draw[0])
        event.Skip()

    def OnWheel(self, event):
        """Mousewheel zoom"""
        if event.WheelRotation < 0:
            direction = 1.0
        else:
            direction = -1.0
        self.canvas.zoom(direction)

    def key_handler(self, event):
        """Keyboard zoom"""
        key = event.GetKeyCode()
        if key in (43, wx.WXK_NUMPAD_ADD):
            # + key zoom in
            self.canvas.zoom(-1)
        elif key in (45, wx.WXK_NUMPAD_SUBTRACT):
            # - key zoom out
            self.canvas.zoom(1)
        else:
            event.Skip()

    def draw(self, objects, title="", xlabel=" ", ylabel=" "):
        """Draw objects to plot"""
        graphics = plot.PlotGraphics(objects, title, xlabel, ylabel)
        self.canvas.Draw(graphics, self.canvas.axis_x, self.canvas.axis_y)
        if self.is_ccss:
            self.canvas.OnMouseDoubleClick(None)

    def draw_ccxx(self):
        """Spectra or matrix 'flower' plot"""
        self.canvas.SetEnableLegend(False)
        self.canvas.proportional = not self.is_ccss
        self.canvas.axis_x = self.ccxx_axis_x
        self.canvas.axis_y = self.ccxx_axis_y
        if self.is_ccss:
            self.canvas.spec_x = self.spec_x
            self.canvas.spec_y = self.spec_y
            self.canvas.SetXSpec(self.spec_x)
            self.canvas.SetYSpec(self.spec_y)
        self.draw(self.gfx, " ")

    def draw_cie(self):
        """CIE 1931 2° xy plot"""
        self.canvas.SetEnableLegend(True)
        self.canvas.proportional = True
        gfx = []
        # Add a few points at the extremes to define a bounding box
        gfx.append(plot.PolyLine([(0, -0.025), (1, 1)], colour=wx.Colour(0, 0, 0, 0)))
        # Add CIE 1931 outline
        gfx.append(
            plot.PolySpline(
                colormath.cie1931_2_xy, colour=wx.Colour(102, 102, 102, 153), width=1.75
            )
        )
        gfx.append(
            plot.PolyLine(
                [colormath.cie1931_2_xy[0], colormath.cie1931_2_xy[-1]],
                colour=wx.Colour(102, 102, 102, 153),
                width=1.75,
            )
        )
        # Add comparison gamuts
        for rgb_space, pen_style in [
            ("Rec. 2020", wx.SOLID),
            ("Adobe RGB (1998)", wx.SHORT_DASH),
            ("DCI P3", wx.DOT_DASH),
            ("Rec. 709", wx.DOT),
        ]:
            values = []
            for R, G, B in [(1, 0, 0), (0, 1, 0), (0, 0, 1)]:
                values.append(colormath.RGB2xyY(R, G, B, rgb_space)[:2])
            values.append(values[0])
            gfx.append(
                plot.PolyLine(
                    values,
                    colour=wx.Colour(102, 102, 102, 255),
                    legend=rgb_space.replace(" (1998)", ""),
                    width=2,
                    style=pen_style,
                )
            )
        # Add points
        for i, (XYZ, _values, _attrs) in enumerate(self.samples):
            if len(XYZ) != 3:
                continue
            xy = colormath.XYZ2xyY(*XYZ)[:2]
            gfx.append(
                plot.PolyMarker(
                    [colormath.XYZ2xyY(*XYZ)[:2]],
                    colour=wx.Colour(*self.gfx[i].attributes["colour"]),
                    size=2,
                    width=1.75,
                    marker="plus",
                    legend="%.4f\u2009x\u2002%.4f\u2009y" % xy,
                )
            )
        self.canvas.axis_x = 0, 1
        self.canvas.axis_y = 0, 1
        self.canvas.spec_x = 10
        self.canvas.spec_y = 10
        self.canvas.SetXSpec(10)
        self.canvas.SetYSpec(10)
        self.draw(gfx, " ", "x", "y")

    def toggle_draw(self, event):
        """Toggle between spectral and CIE plot"""
        if self.canvas.GetEnableLegend():
            self.draw_ccxx()
            self.toggle_btn.SetLabel(lang.getstr("spectral"))
        else:
            self.draw_cie()
            self.toggle_btn.SetLabel(lang.getstr("whitepoint.xy"))



# -*- coding: utf-8 -*-

from io import StringIO
from subprocess import call
from os.path import basename, splitext
import os
import shutil
import sys
import traceback

from DisplayCAL.meta import name
from DisplayCAL.util_os import relpath, safe_glob, which

recordfile_name = "INSTALLED_FILES"

if sys.stdout and hasattr(sys.stdout, "isatty") and not sys.stdout.isatty():
    sys.stdout = StringIO()

if sys.platform == "win32":
    try:
        create_shortcut
    # this function is only available within bdist_wininst installers
    except NameError:
        try:
            from pythoncom import (
                CoCreateInstance,
                CLSCTX_INPROC_SERVER,
                IID_IPersistFile,
            )
            from win32com.shell import shell
            import win32con
        except ImportError:

            def create_shortcut(*args):
                pass

        else:

            def create_shortcut(*args):
                shortcut = CoCreateInstance(
                    shell.CLSID_ShellLink,
                    None,
                    CLSCTX_INPROC_SERVER,
                    shell.IID_IShellLink,
                )
                shortcut.SetPath(args[0])
                shortcut.SetDescription(args[1])
                if len(args) > 3:
                    shortcut.SetArguments(args[3])
                if len(args) > 4:
                    shortcut.SetWorkingDirectory(args[4])
                if len(args) > 5:
                    shortcut.SetIconLocation(args[5], args[6] if len(args) > 6 else 0)
                shortcut.SetShowCmd(win32con.SW_SHOWNORMAL)
                shortcut.QueryInterface(IID_IPersistFile).Save(args[2], 0)

    try:
        directory_created
    # this function is only available within bdist_wininst installers
    except NameError:

        def directory_created(path):
            pass

    try:
        file_created
    # this function is only available within bdist_wininst installers
    except NameError:
        try:
            import win32api
        except ImportError:

            def file_created(path):
                pass

        else:

            def file_created(path):
                if os.path.exists(recordfile_name):
                    installed_files = []
                    if os.path.exists(recordfile_name):
                        recordfile = open(recordfile_name, "r")
                        installed_files.extend(line.rstrip("\n") for line in recordfile)
                        recordfile.close()
                    try:
                        path.encode("ASCII")
                    except (UnicodeDecodeError, UnicodeEncodeError):
                        # the contents of the record file used by distutils
                        # must be ASCII GetShortPathName allows us to avoid
                        # any issues with encoding because it returns the
                        # short path as 7-bit string (while still being a
                        # valid path)
                        path = win32api.GetShortPathName(path)
                    installed_files.append(path)
                    recordfile = open(recordfile_name, "w")
                    recordfile.write("\n".join(installed_files))
                    recordfile.close()

    try:
        get_special_folder_path
    # this function is only available within bdist_wininst installers
    except NameError:
        try:
            from win32com.shell import shell, shellcon
        except ImportError:

            def get_special_folder_path(csidl_string):
                pass

        else:

            def get_special_folder_path(csidl_string):
                return shell.SHGetSpecialFolderPath(
                    0, getattr(shellcon, csidl_string), 1
                )


def postinstall(prefix=None):
    if sys.platform == "darwin":
        # TODO: implement
        pass
    elif sys.platform == "win32":
        if prefix is None:
            # assume we are running from bdist_wininst installer
            modpath = os.path.dirname(os.path.abspath(__file__))
        else:
            # assume we are running from source dir,
            # or from install dir
            modpath = prefix
        if os.path.exists(modpath):
            mainicon = os.path.join(modpath, "theme", "icons", name + ".ico")
            if os.path.exists(mainicon):
                try:
                    startmenu_programs_common = get_special_folder_path(
                        "CSIDL_COMMON_PROGRAMS"
                    )
                    startmenu_programs = get_special_folder_path("CSIDL_PROGRAMS")
                    startmenu_common = get_special_folder_path("CSIDL_COMMON_STARTMENU")
                    startmenu = get_special_folder_path("CSIDL_STARTMENU")
                except OSError:
                    traceback.print_exc()
                    return
                else:
                    filenames = [
                        filename
                        for filename in safe_glob(
                            os.path.join(sys.prefix, "Scripts", name + "*")
                        )
                        if not filename.endswith("-script.py")
                        and not filename.endswith("-script.pyw")
                        and not filename.endswith(".manifest")
                        and not filename.endswith(".pyc")
                        and not filename.endswith(".pyo")
                        and not filename.endswith("_postinstall.py")
                    ] + ["LICENSE.txt", "README.html", "Uninstall"]
                    installed_shortcuts = []
                    for path in (startmenu_programs_common, startmenu_programs):
                        if path:
                            grppath = os.path.join(path, name)
                            if path == startmenu_programs:
                                group = relpath(grppath, startmenu)
                            else:
                                group = relpath(grppath, startmenu_common)
                            if not os.path.exists(grppath):
                                try:
                                    os.makedirs(grppath)
                                except Exception:
                                    # maybe insufficient privileges?
                                    pass
                            if os.path.exists(grppath):
                                print(
                                    ("Created start menu group '%s' in " "%s")
                                    % (
                                        name,
                                        (
                                            str(path, "MBCS", "replace")
                                            if type(path) != str
                                            else path
                                        ).encode("MBCS", "replace"),
                                    )
                                )
                            else:
                                print(
                                    ("Failed to create start menu group '%s' in " "%s")
                                    % (
                                        name,
                                        (
                                            str(path, "MBCS", "replace")
                                            if type(path) != str
                                            else path
                                        ).encode("MBCS", "replace"),
                                    )
                                )
                                continue
                            directory_created(grppath)
                            for filename in filenames:
                                lnkname = splitext(basename(filename))[0]
                                lnkpath = os.path.join(grppath, lnkname + ".lnk")
                                if os.path.exists(lnkpath):
                                    try:
                                        os.remove(lnkpath)
                                    except Exception:
                                        # maybe insufficient privileges?
                                        print(
                                            (
                                                "Failed to create start menu entry '%s' in "
                                                "%s"
                                            )
                                            % (
                                                lnkname,
                                                (
                                                    str(grppath, "MBCS", "replace")
                                                    if type(grppath) != str
                                                    else grppath
                                                ).encode("MBCS", "replace"),
                                            )
                                        )
                                        continue
                                if not os.path.exists(lnkpath):
                                    if lnkname != "Uninstall":
                                        tgtpath = os.path.join(modpath, filename)
                                    try:
                                        if lnkname == "Uninstall":
                                            uninstaller = os.path.join(
                                                sys.prefix, "Remove%s.exe" % name
                                            )
                                            if os.path.exists(uninstaller):
                                                create_shortcut(
                                                    uninstaller,
                                                    lnkname,
                                                    lnkpath,
                                                    '-u "%s-wininst.log"'
                                                    % os.path.join(sys.prefix, name),
                                                    sys.prefix,
                                                    os.path.join(
                                                        modpath,
                                                        "theme",
                                                        "icons",
                                                        name + "-uninstall.ico",
                                                    ),
                                                )
                                            else:
                                                # When running from a
                                                # bdist_wininst or bdist_msi
                                                # installer, sys.executable
                                                # points to the installer
                                                # executable, not python.exe
                                                create_shortcut(
                                                    os.path.join(
                                                        sys.prefix, "python.exe"
                                                    ),
                                                    lnkname,
                                                    lnkpath,
                                                    '"%s" uninstall '
                                                    '--record="%s"'
                                                    % (
                                                        os.path.join(
                                                            modpath, "setup.py"
                                                        ),
                                                        os.path.join(
                                                            modpath, "INSTALLED_FILES"
                                                        ),
                                                    ),
                                                    sys.prefix,
                                                    os.path.join(
                                                        modpath,
                                                        "theme",
                                                        "icons",
                                                        name + "-uninstall.ico",
                                                    ),
                                                )
                                        elif lnkname.startswith(name):
                                            # When running from a
                                            # bdist_wininst or bdist_msi
                                            # installer, sys.executable
                                            # points to the installer
                                            # executable, not python.exe
                                            icon = os.path.join(
                                                modpath,
                                                "theme",
                                                "icons",
                                                lnkname + ".ico",
                                            )
                                            if not os.path.isfile(icon):
                                                icon = mainicon
                                            if filename.endswith(".exe"):
                                                exe = filename
                                                args = ""
                                            else:
                                                exe = os.path.join(
                                                    sys.prefix, "pythonw.exe"
                                                )
                                                args = '"%s"' % tgtpath
                                            create_shortcut(
                                                exe,
                                                lnkname,
                                                lnkpath,
                                                args,
                                                modpath,
                                                icon,
                                            )
                                        else:
                                            create_shortcut(
                                                tgtpath, lnkname, lnkpath, "", modpath
                                            )
                                    except Exception:
                                        # maybe insufficient privileges?
                                        print(
                                            (
                                                "Failed to create start menu entry '%s' in "
                                                "%s"
                                            )
                                            % (
                                                lnkname,
                                                (
                                                    str(grppath, "MBCS", "replace")
                                                    if type(grppath) != str
                                                    else grppath
                                                ).encode("MBCS", "replace"),
                                            )
                                        )
                                        continue
                                    print(
                                        ("Installed start menu entry '%s' to " "%s")
                                        % (
                                            lnkname,
                                            (
                                                str(group, "MBCS", "replace")
                                                if type(group) != str
                                                else group
                                            ).encode("MBCS", "replace"),
                                        )
                                    )
                                file_created(lnkpath)
                                installed_shortcuts.append(filename)
                            if installed_shortcuts == filenames:
                                break
            else:
                print("warning - '%s' not found" % icon.encode("MBCS", "replace"))
            if os.path.exists(recordfile_name):
                irecordfile_name = os.path.join(modpath, "INSTALLED_FILES")
                irecordfile = open(irecordfile_name, "w")
                irecordfile.close()
                file_created(irecordfile_name)
                shutil.copy2(recordfile_name, irecordfile_name)
        else:
            print("warning - '%s' not found" % modpath.encode("MBCS", "replace"))
    else:
        # Linux/Unix
        if prefix is None:
            prefix = sys.prefix
        if which("touch"):
            call(["touch", "--no-create", prefix + "/share/icons/hicolor"])
        if which("xdg-icon-resource"):
            # print "installing icon resources..."
            # for size in [16, 22, 24, 32, 48, 256]:
            # call(["xdg-icon-resource", "install", "--noupdate", "--novendor",
            # "--size", str(size), prefix +
            # ("/share/%s/theme/icons/%sx%s/%s.png" % (name, size, size,
            # name))])
            call(["xdg-icon-resource", "forceupdate"])
        if which("xdg-desktop-menu"):
            # print "installing desktop menu entry..."
            # call(["xdg-desktop-menu", "install", "--novendor", (prefix +
            # "/share/%s/%s.desktop" % (name, name))])
            call(["xdg-desktop-menu", "forceupdate"])


def postuninstall(prefix=None):
    if sys.platform == "darwin":
        # TODO: implement
        pass
    elif sys.platform == "win32":
        # nothing to do
        pass
    else:
        # Linux/Unix
        if prefix is None:
            prefix = sys.prefix
        if which("xdg-desktop-menu"):
            # print "uninstalling desktop menu entry..."
            # call(["xdg-desktop-menu", "uninstall", prefix +
            # ("/share/applications/%s.desktop" % name)])
            call(["xdg-desktop-menu", "forceupdate"])
        if which("xdg-icon-resource"):
            # print "uninstalling icon resources..."
            # for size in [16, 22, 24, 32, 48, 256]:
            # call(["xdg-icon-resource", "uninstall", "--noupdate", "--size",
            # str(size), name])
            call(["xdg-icon-resource", "forceupdate"])


def main():
    prefix = None
    for arg in sys.argv[1:]:
        arg = arg.split("=")
        if len(arg) == 2:
            if arg[0] == "--prefix":
                prefix = arg[1]
    try:
        if "-remove" in sys.argv[1:]:
            postuninstall(prefix)
        else:
            postinstall(prefix)
    except Exception:
        traceback.print_exc()


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

# Argyll CMS tools used by DisplayCAL
names = [
    "applycal",
    "average",
    "cctiff",
    "ccxxmake",
    "dispcal",
    "dispread",
    "collink",
    "colprof",
    "dispwin",
    "fakeread",
    "iccgamut",
    "icclu",
    "xicclu",
    "spotread",
    "spyd2en",
    "spyd4en",
    "targen",
    "tiffgamut",
    "timage",
    "txt2ti3",
    "i1d3ccss",
    "viewgam",
    "oeminst",
    "profcheck",
    "spec2cie",
]

# Argyll CMS tools optionally used by DisplayCAL
optional = [
    "applycal",
    "average",
    "cctiff",
    "ccxxmake",
    "i1d3ccss",
    "oeminst",
    "spec2cie",
    "spyd2en",
    "spyd4en",
    "tiffgamut",
    "timage",
]

prefixes_suffixes = ["argyll"]

# Alternative tool names (from older Argyll CMS versions or with filename
# prefix/suffix like on some Linux distros)
altnames = {
    "txt2ti3": ["logo2cgats"],
    "icclu": ["xicclu"],
    "ccxxmake": ["ccmxmake"],
    "i1d3ccss": ["oeminst"],
    "spyd2en": ["oeminst"],
    "spyd4en": ["oeminst"],
}


def add_prefixes_suffixes(name, altname):
    for prefix_suffix in prefixes_suffixes:
        altnames[name].append(f"{altname}-{prefix_suffix}")
        altnames[name].append(f"{prefix_suffix}-{altname}")


# Automatically populate the alternative tool names with prefixed/suffixed
# versions
for name in names:
    if name not in altnames:
        altnames[name] = []
    _altnames = list(altnames[name])
    for altname in _altnames:
        add_prefixes_suffixes(name, altname)
    altnames[name].append(name)
    add_prefixes_suffixes(name, name)
    altnames[name].reverse()

# Viewing conditions supported by colprof (only predefined choices)
viewconds = [
    "pp",
    "pe",
    "pc",  # Argyll 1.1.1
    "mt",
    "mb",
    "md",
    "jm",
    "jd",
    "tv",  # Argyll 1.6
    "pcd",
    "ob",
    "cx",
]

# Intents supported by colprof
# pa = Argyll >= 1.3.3
# lp = Argyll >= 1.8.3
intents = ["a", "aa", "aw", "la", "lp", "ms", "p", "pa", "r", "s"]

# Video input/output encodings supported by collink (Argyll >= 1.6)
video_encodings = ["n", "t", "6", "7", "5", "2", "C", "x", "X"]

# Observers
observers = ["1931_2", "1955_2", "1964_10", "1964_10c", "1978_2", "shaw"]



# Common py2exe boot script - executed for all target types.

# When we are a windows_exe we have no console, and writing to
# sys.stderr or sys.stdout will sooner or later raise an exception,
# and tracebacks will be lost anyway (see explanation below).
#
# We assume that output to sys.stdout can go to the bitsink, but we
# *want* to see tracebacks.  So we redirect sys.stdout into an object
# with a write method doing nothing, and sys.stderr into a logfile
# having the same name as the executable, with '.log' appended.
#
# We only open the logfile if something is written to sys.stderr.
#
# If the logfile cannot be opened for *any* reason, we have no choice
# but silently ignore the error.
#
# More elaborate explanation on why this is needed:
#
# The sys.stdout and sys.stderr that GUI programs get (from Windows) are
# more than useless.  This is not a py2exe problem, pythonw.exe behaves
# in the same way.
#
# To demonstrate, run this program with pythonw.exe:
#
# import sys
# sys.stderr = open("out.log", "w")
# for i in range(10000):
#     print i
#
# and open the 'out.log' file.  It contains this:
#
# Traceback (most recent call last):
#   File "out.py", line 6, in ?
#     print i
# IOError: [Errno 9] Bad file descriptor
#
# In other words, after printing a certain number of bytes to the
# system-supplied sys.stdout (or sys.stderr) an exception will be raised.
#

import sys
if sys.frozen == "windows_exe":
    class Stderr(object):
        encoding = None
        softspace = 0
        _file = None
        _error = None

        def write(self, text, alert=sys._MessageBox, fname=sys.executable + '.log'):
            if self._file is None and self._error is None:
                try:
                    self._file = open(fname, 'w')
                except Exception as details:
                    self._error = details
            if self._file is not None:
                self._file.write(text)
                self._file.flush()

        def flush(self):
            if self._file is not None:
                self._file.flush()

        def isatty(self):
            return False

    sys.stderr = Stderr()
    del sys._MessageBox
    del Stderr

    class Blackhole(object):
        encoding = None
        softspace = 0

        def write(self, text):
            pass

        def flush(self):
            pass

        def isatty(self):
            return False

    sys.stdout = Blackhole()
    del Blackhole

del sys

# Disable linecache.getline() which is called by
# traceback.extract_stack() when an exception occurs to try and read
# the filenames embedded in the packaged python code.  This is really
# annoying on windows when the d: or e: on our build box refers to
# someone elses removable or network drive so the getline() call
# causes it to ask them to insert a disk in that drive.
import linecache
def fake_getline(filename, lineno, module_globals=None):
    return ''
linecache.orig_getline = linecache.getline
linecache.getline = fake_getline

del linecache, fake_getline



# -*- coding: utf-8 -*-

import os
import sys
import tempfile

sys.path.insert(
    1,
    os.path.join(
        os.path.dirname(os.path.dirname(os.path.abspath(__file__))), "DisplayCAL"
    ),
)


from DisplayCAL.meta import author, description, DOMAIN, name, version, version_tuple


def mktempver(
    version_template_path, name_=name, description_=description, encoding="UTF-8"
):
    version_template = open(version_template_path, "rb")
    tempver_str = version_template.read().decode(encoding, "replace") % {
        "filevers": str(version_tuple),
        "prodvers": str(version_tuple),
        "CompanyName": DOMAIN,
        "FileDescription": description_,
        "FileVersion": f"{version}",
        "InternalName": name_,
        "LegalCopyright": f"© {author}",
        "OriginalFilename": f"{name_}.exe",
        "ProductName": name_,
        "ProductVersion": f"{version}",
    }
    version_template.close()
    tempdir = tempfile.mkdtemp()
    tempver_path = os.path.join(tempdir, "winversion.txt")
    if not os.path.exists(tempdir):
        os.makedirs(tempdir)
    tempver = open(tempver_path, "wb")
    tempver.write(tempver_str.encode(encoding, "replace"))
    tempver.close()
    return tempver_path



# -*- coding: utf-8 -*-
from __future__ import annotations

import io
import pathlib
import os
import shutil
import sys
import tempfile
from typing import Tuple, Dict
from urllib.error import URLError

import pytest

from DisplayCAL import ICCProfile, CGATS
from DisplayCAL import config
from DisplayCAL.dev.mocks import check_call_str
from DisplayCAL.worker import (
    get_argyll_version_string,
    make_argyll_compatible_path,
    Worker,
    add_keywords_to_cgats,
    check_create_dir,
    check_cal_isfile,
    check_profile_isfile,
    check_file_isfile,
    check_ti3_criteria1,
    get_argyll_latest_version,
)

from DisplayCAL.worker_base import get_argyll_util
from tests.data.display_data import DisplayData


def test_get_options_from_profile_1(data_files):
    """Test ``DisplayCAL.worker.get_options_from_profile()`` function"""
    from DisplayCAL.worker import get_options_from_profile

    profile_path = data_files[
        "UP2516D #1 2022-03-23 16-06 D6500 2.2 F-S XYZLUT+MTX.icc"
    ].absolute()
    options = get_options_from_profile(profile=profile_path)
    assert options == (
        [
            "t6500",
            "g2.2",
            "f1.0",
            "A4.0",
            "d1",
            "c1",
            "yl",
            "P0.48923385077616427,0.8797619047619047,1.4894179894179895",
            "H",
        ],
        ["qh", "aX", 'A "Dell, Inc."'],
    )


def test_get_options_from_profile_2(data_files):
    """Test ``DisplayCAL.worker.get_options_from_profile()`` function, for #69"""
    from DisplayCAL.worker import get_options_from_profile

    profile_path = data_files["SW271 PM PenalNative_KB1_160_2022-03-17.icc"].absolute()
    options = get_options_from_profile(profile=profile_path)
    assert options == ([], [])  # no options on that profile


def test_make_argyll_compatible_path_1():
    """testing if make_argyll_compatible_path is working properly with bytes input"""
    test_value = "C:\\Program Files\\some path\\excutable.exe"
    result = make_argyll_compatible_path(test_value)
    expected_result = "C_Program Files_some path_excutable.exe"
    assert result == expected_result


def test_make_argyll_compatible_path_2():
    """testing if make_argyll_compatible_path is working properly with bytes input"""
    test_value = b"C:\\Program Files\\some path\\excutable.exe"
    result = make_argyll_compatible_path(test_value)
    expected_result = b"C_Program Files_some path_excutable.exe"
    assert result == expected_result


def test_worker_get_instrument_name_1():
    """testing if the Worker.get_instrument_name() is working properly"""
    worker = Worker()
    result = worker.get_instrument_name()
    expected_result = ""
    assert result == expected_result


def test_worker_get_instrument_features():
    """testing if Worker.get_instrument_features() is working properly"""
    worker = Worker()
    result = worker.get_instrument_features()
    assert result == {}


def test_worker_instrument_supports_css_1():
    """testing if Worker.instrument_supports_ccss is working properly"""
    worker = Worker()
    result = worker.instrument_supports_ccss()
    expected_result = None
    assert result == expected_result


# @pytest.mark.skip(reason="Test segfaults with python 3.12 - further investigation required.")
def test_generate_b2a_from_inverse_table(data_files, argyll):
    """Test Worker.generate_B2A_from_inverse_table() method"""
    worker = Worker()
    icc_profile1 = ICCProfile.ICCProfile(
        profile=data_files[
            "Monitor 1 #1 2022-03-09 16-13 D6500 2.2 F-S XYZLUT+MTX.icc"
        ].absolute()
    )
    logfile = io.StringIO()
    result = worker.generate_B2A_from_inverse_table(icc_profile1, logfile=logfile)
    assert result is True


def test_get_argyll_version_1(argyll):
    """Test worker.get_argyll_version() function."""
    from DisplayCAL.worker import get_argyll_version

    with check_call_str("DisplayCAL.worker.base_get_argyll_version_string", "2.3.0"):
        result = get_argyll_version("ccxxmake")
    expected_result = [2, 3, 0]
    assert result == expected_result


def test_sudo_class_initialization():
    """Test worker.Sudo class initialization"""
    from DisplayCAL.worker import Sudo

    sudo = Sudo()
    assert sudo is not None


def test_download_method_1():
    """Test Worker.download() method."""
    from DisplayCAL.meta import DOMAIN
    from DisplayCAL.worker import Worker

    worker = Worker()
    uri = f"https://{DOMAIN}/i1d3"
    result = worker.download(uri)
    assert result is not None


def test_download_method_2():
    """Test Worker.download() method."""
    from DisplayCAL.meta import DOMAIN
    from DisplayCAL.worker import Worker

    worker = Worker()
    uri = f"https://{DOMAIN}/i1d3"
    result = worker.download(uri, force=True)
    assert result is not None


def test_download_method_3():
    """Test Worker.download() method."""
    from DisplayCAL.meta import DOMAIN
    from DisplayCAL.worker import Worker

    worker = Worker()
    uri = f"https://{DOMAIN}/spyd2"
    result = worker.download(uri)
    assert result is not None


def test_download_method_4():
    """Test Worker.download() method."""
    from DisplayCAL.meta import DOMAIN
    from DisplayCAL.worker import Worker

    worker = Worker()
    uri = f"https://{DOMAIN}/spyd2"
    result = worker.download(uri, force=True)
    assert result is not None


def test_get_display_name_1():
    """Testing Worker.get_display_name() method."""
    from DisplayCAL.worker import Worker
    from DisplayCAL.config import initcfg, setcfg

    initcfg()
    setcfg("display.number", 1)
    worker = Worker()
    result = worker.get_display_name(False, True, False)
    assert result == ""


def test_get_pwd():
    """Testing Worker.get_display_name() method."""
    from DisplayCAL.worker import Worker
    from DisplayCAL.config import initcfg

    initcfg()
    worker = Worker()
    test_value = "test_value"
    worker.pwd = test_value
    assert worker.pwd == test_value


def test_update_profile_1(random_icc_profile):
    """Testing Worker.update_profile() method."""
    from DisplayCAL import worker
    from DisplayCAL.worker import Worker
    from DisplayCAL.config import initcfg

    worker.dbus_session = None
    worker.dbus_system = None
    initcfg()
    worker = Worker()

    icc_profile, icc_profile_path = random_icc_profile
    with check_call_str(
        "DisplayCAL.worker.Worker.get_display_edid", DisplayData.DISPLAY_DATA_2
    ):
        worker.update_profile(icc_profile_path, tags=True)


def test_exec_cmd_1():
    """Test worker.exec_cmd() function for issue #73"""
    # Command line:
    from DisplayCAL.worker import Worker

    cmd = "/home/eoyilmaz/.local/bin/Argyll_V2.3.0/bin/colprof"
    args = [
        "-v",
        "-qh",
        "-ax",
        "-bn",
        "-C",
        b"No copyright. Created with DisplayCAL 3.8.9.3 and Argyll CMS 2.3.0",
        "-A",
        "Dell, Inc.",
        "-D",
        "UP2516D_#1_2022-04-01_00-26_2.2_F-S_XYZLUT+MTX",
        "/tmp/DisplayCAL-i91d9z8_/UP2516D_#1_2022-04-01_00-26_2.2_F-S_XYZLUT+MTX",
    ]
    cwd = "/tmp/DisplayCAL-i91d9z8_"
    worker = Worker()
    worker.exec_cmd(cmd=cmd, args=args)


def test_is_allowed_1():
    """Test Sudo.is_allowed() function for issue #76"""
    from DisplayCAL.worker import Sudo

    sudo = Sudo()
    result = sudo.is_allowed()
    assert result != ""


def test_ti3_lookup_to_ti1_1(data_files):
    """Test Worker.ti3_lookup_to_ti1() function for #129"""
    ti3_path = data_files["0_16_from_issue_129.ti3"].absolute()
    profile_path = data_files[
        "UP2516D #1 2022-03-23 16-06 D6500 2.2 F-S XYZLUT+MTX.icc"
    ].absolute()
    from DisplayCAL import config, ICCProfile

    ti3_cgat = CGATS.CGATS(ti3_path)
    icc_profile = ICCProfile.ICCProfile(profile_path)
    config.initcfg()
    worker = Worker()
    ti1, ti3v = worker.ti3_lookup_to_ti1(ti3_cgat, icc_profile)
    assert isinstance(ti1, CGATS.CGATS)
    assert isinstance(ti3v, CGATS.CGATS)
    assert ti1 == {
        0: {
            "COLOR_REP": b"RGB",
            "DATA": {
                0: {
                    "RGB_B": 99.9959,
                    "RGB_G": 100.0,
                    "RGB_R": 97.1526,
                    "SAMPLE_ID": 1,
                    "XYZ_X": 95.0104,
                    "XYZ_Y": 100.0,
                    "XYZ_Z": 92.7202,
                },
                1: {
                    "RGB_B": 9.1428,
                    "RGB_G": 5.8338,
                    "RGB_R": 5.842,
                    "SAMPLE_ID": 2,
                    "XYZ_X": 0.277593,
                    "XYZ_Y": 0.255279,
                    "XYZ_Z": 0.423145,
                },
                2: {
                    "RGB_B": 11.6181,
                    "RGB_G": 9.1081,
                    "RGB_R": 8.0801,
                    "SAMPLE_ID": 3,
                    "XYZ_X": 0.51238,
                    "XYZ_Y": 0.536117,
                    "XYZ_Z": 0.705578,
                },
            },
            "DATA_FORMAT": {
                0: b"SAMPLE_ID",
                1: b"RGB_R",
                2: b"RGB_G",
                3: b"RGB_B",
                4: b"XYZ_X",
                5: b"XYZ_Y",
                6: b"XYZ_Z",
            },
            "DESCRIPTOR": b"Argyll Calibration Target chart information 1",
            "KEYWORDS": {0: b"COLOR_REP"},
            "NUMBER_OF_FIELDS": None,
            "NUMBER_OF_SETS": None,
        }
    }

    assert ti3v == {
        "COLOR_REP": b"RGB_XYZ",
        "CREATED": b"Sun Jun  5 13:08:54 2022",
        "DATA": {
            0: {
                "RGB_B": 99.9959,
                "RGB_G": 100.0,
                "RGB_R": 97.1526,
                "SAMPLE_ID": 1,
                "XYZ_X": 95.0104,
                "XYZ_Y": 100.0,
                "XYZ_Z": 92.7202,
            },
            1: {
                "RGB_B": 9.1428,
                "RGB_G": 5.8338,
                "RGB_R": 5.842,
                "SAMPLE_ID": 2,
                "XYZ_X": 0.277593,
                "XYZ_Y": 0.255279,
                "XYZ_Z": 0.423145,
            },
            2: {
                "RGB_B": 11.6181,
                "RGB_G": 9.1081,
                "RGB_R": 8.0801,
                "SAMPLE_ID": 3,
                "XYZ_X": 0.51238,
                "XYZ_Y": 0.536117,
                "XYZ_Z": 0.705578,
            },
        },
        "DATA_FORMAT": {
            0: b"SAMPLE_ID",
            1: b"RGB_R",
            2: b"RGB_G",
            3: b"RGB_B",
            4: b"XYZ_X",
            5: b"XYZ_Y",
            6: b"XYZ_Z",
        },
        "DESCRIPTOR": b"Argyll Calibration Target chart information 3",
        "DEVICE_CLASS": b"DISPLAY",
        "DISPLAY_TYPE_BASE_ID": 1,
        "DISPLAY_TYPE_REFRESH": b"NO",
        "INSTRUMENT_TYPE_SPECTRAL": b"NO",
        "LUMINANCE_XYZ_CDM2": b"42.204124 44.420532 41.186805",
        "NORMALIZED_TO_Y_100": b"YES",
        "NUMBER_OF_FIELDS": None,
        "NUMBER_OF_SETS": None,
        "ORIGINATOR": b"Argyll dispread",
        "TARGET_INSTRUMENT": b"Datacolor Spyder3",
        "VIDEO_LUT_CALIBRATION_POSSIBLE": b"YES",
    }


def test_add_keywords_to_cgats(data_files) -> None:
    """Test if keywords are added to cgats by add_keywords_to_cgats."""
    path = data_files["0_16.ti3"].absolute()
    cgats = CGATS.CGATS(cgats=path)
    assert "keyword" not in cgats[0]
    options = {"keyword": "Value"}
    alternated_cgats = add_keywords_to_cgats(cgats, options)
    assert "keyword" in alternated_cgats[0]


def test_check_create_dir() -> None:
    """Test function 'check_create_dir'."""
    assert check_create_dir("test_dir") == True


@pytest.mark.parametrize("file", (True, False))
def test_check_cal_isfile(data_files, file: bool) -> None:
    """Test 'check_cal_isfile'."""
    path = data_files["Monitor.cal"].absolute() if file else "no_file"
    assert check_cal_isfile(path) == True if file else "error.calibration.file_missing"


@pytest.mark.parametrize("file", (True, False))
def test_check_profile_isfile(data_files, file: bool) -> None:
    """Test 'check_profile_isfile'."""
    path = data_files["Monitor.cal"].absolute() if file else "no_file"
    assert check_profile_isfile(path) == True if file else "error.profile.file_missing"


# todo: test is working locally but not on CI
@pytest.mark.skip(
    reason="First execution of test fails on remote CI server. "
    "All following tests are positive."
)
@pytest.mark.parametrize("silent", (True, False), ids=("silent", "not silent"))
@pytest.mark.parametrize(
    "path,result",
    (
        ("data/cgats0.txt", ("True", "True")),
        ("no_file", ("False", "file.missing")),
        (".", ("False", "file.notfile")),
    ),
)
def test_check_file_isfile(
    data_files, silent: bool, path: str, result: Tuple[str, str]
) -> None:
    """Test if file gets detected."""
    assert (
        str(check_file_isfile(path, silent=silent)) == result[0]
        if silent
        else result[1]
    )


@pytest.mark.parametrize(
    "sample,result",
    (
        (
            {
                "SAMPLE_ID": 1,
                "RGB_R": 50,
                "RGB_G": 50,
                "RGB_B": 50,
                "XYZ_X": 0.5,
                "XYZ_Y": 0.5,
                "XYZ_Z": 0.5,
            },
            True,
        ),
        (
            {
                "SAMPLE_ID": 2,
                "RGB_R": 6,
                "RGB_G": 6,
                "RGB_B": 6,
                "XYZ_X": 0.5,
                "XYZ_Y": 0.5,
                "XYZ_Z": 0.5,
            },
            False,
        ),
    ),
)
def test_check_ti3_criteria1(sample: Dict[str:float], result: bool) -> None:
    """Test for ti3 criteria1 check."""
    black = (0, 0, 0)
    white = (110, 110, 110)
    criteria = check_ti3_criteria1(
        (sample["RGB_R"], sample["RGB_G"], sample["RGB_B"]),
        (sample["XYZ_X"], sample["XYZ_Y"], sample["XYZ_Z"]),
        black,
        white,
        print_debuginfo=True,
    )
    assert criteria[3] == result


def test_prepare_colprof_for_271(monkeypatch, data_path):
    """Bug report 271."""
    assert isinstance(data_path, pathlib.Path)

    def patched_getcfg(key):
        """patched getcfg()"""
        cfg = {
            "argyll.version": "2.3.1",
            "profile.name.expanded": "test_profile",
            "profile.quality": "m",
            "profile.type": "l",
            "gamap_saturation": False,
            "gamap_perceptual": False,
            "profile.quality.b2a": "h",
            "profile.b2a.hires": True,
            "copyright": "",
            "extra_args.colprof": "",
            "profile.black_point_compensation": False,
            "profile.black_point_correction": False,
            "profile.b2a.hires.size": 17,
            "profile.b2a.hires.smooth": True,
            "measure.override_min_display_update_delay_ms": False,
            "measure.override_display_settle_time_mult": False,
            "patterngenerator.ffp_insertion": False,
            "testchart.patch_sequence": "",
            "3dlut.create": False,
        }
        return cfg[key]

    monkeypatch.setattr("DisplayCAL.worker.getcfg", patched_getcfg)

    def patched_os_path_exists(filepath):
        return True

    monkeypatch.setattr("DisplayCAL.worker.os.path.exists", patched_os_path_exists)

    def patched_os_path_isfile(filepath):
        return True

    monkeypatch.setattr("DisplayCAL.worker.os.path.isfile", patched_os_path_isfile)

    worker = Worker()
    in_out_file = pathlib.Path(worker.setup_inout("test_profile")).with_suffix(".ti3")

    # copy the test file to the target path
    test_file_path = data_path / "sample" / "issue271" / "test_profile.ti3"
    os.makedirs(in_out_file.parent, exist_ok=True)
    shutil.copy(test_file_path, in_out_file)

    # This should not raise the:
    # TypeError: startswith first arg must be bytes or a tuple of bytes, not str
    worker.prepare_colprof()


def test_prepare_dispcal_1():
    """Worker.prepare_dispcal() return value should be quoted properly."""
    worker = Worker()
    return_val = worker.prepare_dispcal()
    expected_result = [
        "-v2",
        "-d0",
        "-c1",
        return_val[1][3],  # '-yl',
        return_val[1][4],  # '-P0.5,0.5,1.0',
        "-ql",
        return_val[1][6],  # '-t',
        "-g2.2",
        "-f1.0",
        return_val[1][9],  # '-k0.0',
        "/var/folders/8l/xy1__ym94nn35x86xyg56xq80000gn/T/DisplayCAL-2fdjtyql/",
    ]
    assert return_val[0] == get_argyll_util("dispcal")
    assert isinstance(return_val[1], list)
    assert return_val[1][:-1] == expected_result[:-1]  # don't check the final part
    assert tempfile.gettempdir() in return_val[1][-1]  # this should be in a temp path


@pytest.mark.skipif(
    os.getenv("GITHUB_ACTIONS") == "true",
    reason="Not working properly on GitHub.",
)
def test_get_argyll_version_string_returns_a_proper_value():
    """get_argyll_version_string() returns a proper value."""
    import wx

    config.initcfg()
    app = wx.GetApp() or wx.App()

    assert "0.0.0" != get_argyll_version_string(name="ccxxmake", silent=False)


def test_get_argyll_latest_version_returns_str():
    """get_argyll_latest_version() returns a str."""
    result = get_argyll_latest_version()
    assert isinstance(result, str)


def test_get_argyll_latest_version_returns_latest_argyll_cms_version():
    """get_argyll_latest_version() returns the latest argyll cms version."""
    result = get_argyll_latest_version()
    assert result == "3.3.0"


def test_get_argyll_latest_version_returns_the_default_version_if_no_internet_connect(
    monkeypatch,
):
    """get_argyll_latest_version() returns the default argyll cms version if no internet connection."""

    def patched_urlopen(*args, **kwargs):
        raise URLError(
            "<urlopen error [Errno 8] nodename nor servname provided, or not known>"
        )

    monkeypatch.setattr("DisplayCAL.worker.urllib.request.urlopen", patched_urlopen)
    result = get_argyll_latest_version()
    assert result == config.defaults.get("argyll.version")



"""
This is the Advanced Generic Widgets package (AGW). It provides many 
custom-drawn wxPython controls: some of them can be used as a replacement 
of the platform native controls, others are simply an addition to the 
already rich wxPython widgets set.


Description:

AGW contains many different modules, listed below. Items labelled with 
an asterisk were already present in `wx.lib` before:

- AdvancedSplash: reproduces the behaviour of `wx.SplashScreen`, with more
  advanced features like custom shapes and text animations;
- AquaButton: this is another custom-drawn button class which
  *approximatively* mimics the behaviour of Aqua buttons on the Mac;
- AUI: a pure-Python implementation of `wx.aui`, with many bug fixes and
  new features like HUD docking and L{AuiNotebook} tab arts;
- BalloonTip: allows you to display tooltips in a balloon style window
  (actually a frame), similarly to the Windows XP balloon help;
- ButtonPanel (*): a panel with gradient background shading with the
  possibility to add buttons and controls still respecting the gradient
  background;
- CubeColourDialog: an alternative implementation of `wx.ColourDialog`, it
  offers different functionalities like colour wheel and RGB cube;
- CustomTreeCtrl (*): mimics the behaviour of `wx.TreeCtrl`, with almost the
  same base functionalities plus a bunch of enhancements and goodies;
- FlatMenu: as the name implies, it is a generic menu implementation,
  offering the same `wx.MenuBar`/`wx.Menu`/`wx.ToolBar` capabilities and much more;
- FlatNotebook (*): a full implementation of the `wx.Notebook`, and designed
  to be a drop-in replacement for `wx.Notebook` with enhanced capabilities;
- FloatSpin: this class implements a floating point spinctrl, cabable (in
  theory) of handling infinite-precision floating point numbers;
- FoldPanelBar (*): a control that contains multiple panels that can be
  expanded or collapsed a la Windows Explorer/Outlook command bars;
- FourWaySplitter: this is a layout manager which manages four children like
  four panes in a window, similar to many CAD software interfaces;
- GenericMessageDialog: it is a possible replacement for the standard
  `wx.MessageDialog`, with a fancier look and extended functionalities;
- GradientButton: another custom-drawn button class which mimics Windows CE
  mobile gradient buttons, using a tri-vertex blended gradient background;
- HyperLinkCtrl (*): this widget acts line an hyper link in a typical browser;
- HyperTreeList: a class that mimics the behaviour of `wx.gizmos.TreeListCtrl`,
  with almost the same base functionalities plus some more enhancements;
- KnobCtrl: a widget which lets the user select a numerical value by
  rotating it, like a slider with a wheel shape;
- LabelBook and FlatImageBook: these are a quasi-full implementations of
  `wx.ListBook`, with additional features;
- MultiDirDialog: it represents a possible replacement for `wx.DirDialog`,
  with the additional ability of selecting multiple folders at once and a
  fancier look;
- PeakMeter: this widget mimics the behaviour of LED equalizers that are
  usually found in stereos and MP3 players;
- PersistentControls: widgets which automatically save their state
  when they are destroyed and restore it when they are recreated, even during
  another program invocation;
- PieCtrl and ProgressPie: these are simple classes that reproduce the
  behavior of a pie chart, in a static or progress-gauge-like way;
- PyBusyInfo: constructs a busy info window and displays a message in it:
  it is similar to `wx.BusyInfo`;
- PyCollapsiblePane: a pure Python implementation of the original wxWidgets
  C++ code of `wx.CollapsiblePane`, with customizable buttons;
- PyGauge: a generic `wx.Gauge` implementation, it supports the determinate
  mode functions as `wx.Gauge`;
- PyProgress: it is similar to `wx.ProgressDialog` in indeterminated mode, but
  with a different gauge appearance and a different spinning behavior;
- RibbonBar: the RibbonBar library is a set of classes for writing a ribbon
  user interface, similar to the user interface present in recent versions
  of Microsoft Office;
- RulerCtrl: it implements a ruler window that can be placed on top, bottom,
  left or right to any wxPython widget. It is somewhat similar to the rulers
  you can find in text editors software;
- ShapedButton: this class tries to fill the lack of "custom shaped" controls
  in wxPython. It can be used to build round buttons or elliptic buttons;
- SpeedMeter: this widget tries to reproduce the behavior of some car
  controls (but not only), by creating an "angular" control;
- SuperToolTip: a class that mimics the behaviour of `wx.TipWindow` and
  generic tooltips, with many features and highly customizable;
- ThumbnailCtrl: a widget that can be used to display a series of images
  in a "thumbnail" format; it mimics, for example, the Windows Explorer
  behavior when you select the "view thumbnails" option;
- ToasterBox: a cross-platform widget to make the creation of MSN-style
  "toaster" popups easier;
- UltimateListCtrl: mimics the behaviour of `wx.ListCtrl`, with almost the same
  base functionalities plus some more enhancements;
- ZoomBar: a class that *appoximatively* mimics the behaviour of the Mac Dock,
  inside a `wx.Panel`.


Bugs and Limitations: many, patches and fixes welcome :-D

See the demos for an example of what AGW can do, and on how to use it.

Copyright: Andrea Gavana

License: Same as the version of wxPython you are using it with.

SVN for latest code:
http://svn.wxwidgets.org/viewvc/wx/wxPython/3rdParty/AGW/

Mailing List:
wxpython-users@lists.wxwidgets.org

My personal web page:
http://xoomer.alice.it/infinity77

Please let me know if you are using AGW!

You can contact me at:

andrea.gavana@gmail.com
gavana@kpo.kz

AGW version: 0.9.1

Last updated: 21 Jun 2011, 22.00 GMT

"""

__version__ = "0.9.1"
__author__ = "Andrea Gavana <andrea.gavana@gmail.com>"



# -*- coding: utf-8 -*-
from typing import Tuple

import pytest
import wx
from _pytest.fixtures import SubRequest

from DisplayCAL import RealDisplaySizeMM
from DisplayCAL.dev.mocks import check_call, check_call_str
from DisplayCAL.wxMeasureFrame import get_default_size
from tests.data.display_data import DisplayData


@pytest.fixture(
    scope="session", name="size_in_mm", params=["size_available", "size_unavailable"]
)
def fixture_size_in_mm(request: SubRequest) -> Tuple[int, int]:
    """Return display size in mm (width, height)."""
    return (
        DisplayData.DISPLAY_DATA_1["size"]
        if request.param == "size_available"
        else (0, 0)
    )


@pytest.mark.parametrize(
    "real_display",
    (True, False),
    ids=("with_RealDisplaySizeMM", "without_RealDisplaySizeMM"),
)
def test_get_default_size_1(real_display: bool, size_in_mm: Tuple[int, int]) -> None:
    """Testing wxMeasureFrame.get_default_size() function."""
    with check_call_str("DisplayCAL.wxMeasureFrame.getcfg", DisplayData.CFG_DATA):
        with check_call_str(
            "DisplayCAL.wxMeasureFrame.get_display_number",
            DisplayData.DISPLAY_DATA_1["screen"],
        ):
            with check_call(wx, "Display", DisplayData()):
                if real_display:
                    with check_call(
                        RealDisplaySizeMM,
                        "RealDisplaySizeMM",
                        DisplayData.DISPLAY_DATA_1["size"],
                    ):
                        result = get_default_size()
                else:
                    with check_call(
                        wx, "DisplaySize", DisplayData.DISPLAY_DATA_1["size"]
                    ):
                        with check_call(wx, "DisplaySizeMM", size_in_mm):
                            result = get_default_size()
    assert isinstance(result, int)
    assert result > 1
